"""
    Copyright (c) 2019 Contributors as noted in the AUTHORS file
    This file is part of ggm, the GILGAMESH core engine in Python.
    ggm is free software; you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License (GPL) as published
    by the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
    As a special exception, the Contributors give you permission to link
    this library with independent modules to produce an executable,
    regardless of the license terms of these independent modules, and to
    copy and distribute the resulting executable under terms of your choice,
    provided that you also meet, for each linked independent module, the
    terms and conditions of the license of that module. An independent
    module is a module which is not derived from or based on this library.
    If you modify this library, you must extend this exception to your
    version of the library.
    ggm is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
    License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import zmq
import socket
import asyncio
from zmq.utils.strtypes import asbytes
from datetime import datetime as dt
import time
 
import random

from ..lib.kernel import GK
from ..lib.kontext import Kontext
from ..lib.auth import GAuthClient

from pprint import pprint

class KUplink(GAuthClient, GK):
    def __init__(self, *args, **kwargs):
        GK.__init__(self, *args, **kwargs)
        # FIXME maybe do this nice or not at all
        self.auth_cfg = {
            'keys_path': kwargs['keys_path'],
            'client_key': kwargs['client_key'],
            'server_key': kwargs['server_key'],
            'hostname': kwargs['hostname'],
            'port': kwargs['port']
        }

        self.FRO = "ipc:///tmp/frontend"
        self.fro = self.context.socket(zmq.DEALER)
        #self.fro.setsockopt(zmq.IDENTITY, asbytes(self.name))
        #self.fro.setsockopt(zmq.LINGER, 0)
        self.fro.connect(self.FRO)

        self.loop.create_task(self.heartbeat())
        self.loop.create_task(self.upload_data())

    async def upload_data(self):
        upl = GAuthClient(**self.auth_cfg)
        while True:
            rand_sleep = round(random.uniform(2.5,4.5), 2)
            await asyncio.sleep(60*rand_sleep)
            #await asyncio.sleep(5)
            # greetings!
            earthling = True
            await upl.client.dsend(['greetings'])
            route, earthling = await upl.client.drecv()
            #self.glog.debug(f'UPLINK got greeting ({earthling}) from {uplink}')

            start_time = time.time()

            # info
            remote_info = {}
            await upl.client.dsend(['info'])
            route, remote_info = await upl.client.drecv()
            #print(route, remote_info)

            # get remote store
            store_remote = {}
            store_query = ['json', 'get', 'device_state_db', remote_info['dev_id'], 'head']
            await upl.client.dsend(store_query)
            route, store_remote = await upl.client.drecv()
            try:
                store_remote = store_remote['device_state_db'][remote_info['dev_id']].pop('store')
                #print(route, store_remote)
            except Exception as e:
                self.glog.error(f'trouble getting remote store on {self.auth_cfg["hostname"]}')
                continue

            # get local state 
            state = {}
            inv_query = ['json', 'get', 'device_state_db', self.dev_id, 'head']
            await self.fro.dsend(inv_query)
            route, state = await self.fro.drecv()
            state = state['device_state_db']

            # update remote state
            upup = ['json', 'nupsert', 'device_state_db', self.dev_id, state[self.dev_id]]
            await upl.client.dsend(upup)
            route, response = await upl.client.drecv()

            #print(self.name)
            #pprint('Local Inventory: ', state)
            #pprint('Remote Store: ', store_remote)
            #self.glog.debug(f'Local Inventory ---- {state}')
            #self.glog.debug(f'Remote Store ---- {store_remote}')

            for col in state[self.dev_id]['inventory']:
                segment = {}
                ## full upload
                try:
                    segment['start'] = state[self.dev_id]['inventory'][col]['tail']['time']
                    segment['stop'] = state[self.dev_id]['inventory'][col]['head']['time']
                except:
                    # FIXME leftovers?
                    continue
                if self.dev_id in store_remote:
                    if col in store_remote[self.dev_id]:
                        ## upload increment
                        if 'head' in store_remote[self.dev_id][col]:
                            segment['start'] = store_remote[self.dev_id][col]['head']['time']
                            segment['stop'] = state[self.dev_id]['inventory'][col]['head']['time']
                            #print('increment')
                #print(segment)
                if segment['start'] == segment['stop']:
                    continue
                cmd = ['series', 'chunk', self.dev_id, col]
                self.glog.info(f'Uploading ({self.name}): collection {col} segment {segment}')
                await upl.client.upstream(cmd, segment, source=self.fro, compression=False)

            self.glog.info(f'Uploaded state snapshot to: {self.auth_cfg["hostname"]}')
            stop_time = time.time()
            self.glog.info(f'Upload took: {round((stop_time-start_time), 3)} s')

    async def heartbeat(self):
        upl = GAuthClient(**self.auth_cfg)
        self.glog.info(f'Starting uplink {self.name}.')
        while True:
            await asyncio.sleep(10)
            # greetings!
            await upl.client.dsend(['greetings'])
            route, earthling = await upl.client.drecv()

            start_time = time.time()
            tmst = dt.utcnow()

            # update remote db
            up_request = ['json', 'nupsert', 'device_state_db', self.dev_id, {'seen': tmst.isoformat(sep=' '), 'dev_id': self.dev_id}]
            await upl.client.dsend(up_request)
            route, raw = await upl.client.drecv()
            dev_db = ['json', 'nupsert', 'device_db', self.dev_id , {'dev_id': self.dev_id, 'version': self.version}]
            await upl.client.dsend(dev_db)
            route, raw = await upl.client.drecv()

            now = dt.utcnow()
            if (now-tmst).seconds > 10: ## 10sec is considered timeout
                self.glog.info("Heart lost at: {0} --- Connection reestablished at {1}".format(dt.fromtimestamp(tmst), dt.fromtimestamp(now)))
            stop_time = time.time()
            self.glog.debug(f'({self.name}: {self.auth_cfg["hostname"]}) Heartbeat round trip: {round((stop_time-start_time)*1000, 1)} ms')

def uplink_process(kcfg):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    kcfg['context'] = Kontext()
    kcfg['loop'] = loop

    upl = KUplink(**kcfg)
    upl.start()



"""
    Copyright (c) 2019 Contributors as noted in the AUTHORS file
    This file is part of ggm, the GILGAMESH core engine in Python.
    ggm is free software; you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License (GPL) as published
    by the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
    As a special exception, the Contributors give you permission to link
    this library with independent modules to produce an executable,
    regardless of the license terms of these independent modules, and to
    copy and distribute the resulting executable under terms of your choice,
    provided that you also meet, for each linked independent module, the
    terms and conditions of the license of that module. An independent
    module is a module which is not derived from or based on this library.
    If you modify this library, you must extend this exception to your
    version of the library.
    ggm is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
    License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import sys
import zmq
import socket
import zmq.auth
from zmq.auth.asyncio import AsyncioAuthenticator

from .kontext import Kontext

from pprint import pprint

class GAuthServer(object):
    """
        TODO: write doc
    """
    def __init__(self, *args, **kwargs):
        from pathlib import Path
        if not kwargs['server_key']:
            raise Exception('Nope!')

        keys_dir = kwargs['keys_path']
        public_keys_dir = keys_dir / 'public_keys'
        secret_keys_dir = keys_dir / 'private_keys'

        if not (keys_dir.exists() and
                public_keys_dir.exists() and
                secret_keys_dir.exists()):
            print("CURVE Server: Certificate missing -- this device needs to be comissioned first!")
            print(public_keys_dir, secret_keys_dir)
            sys.exit(1)

        ctx = kwargs['context']
        # Start an authenticator for this context.
        auth = AsyncioAuthenticator(ctx)
        auth.start()
        #auth.allow('127.0.0.1')
        auth.allow()
        # Tell authenticator to use the certificate in a directory
        auth.configure_curve(domain='*', location=public_keys_dir)

        self.server = ctx.socket(zmq.ROUTER)

        server_secret_file = secret_keys_dir / Path(f'{kwargs["server_key"]}.key_secret')
        server_public, server_secret = zmq.auth.load_certificate(server_secret_file)
        self.server.curve_secretkey = server_secret
        self.server.curve_publickey = server_public
        self.server.curve_server = True  # must come before bind
        self.server.bind(f'tcp://*:{kwargs["port"]}')

class GAuthClient(object):
    """
        TODO: write doc
    """
    def __init__(self, *args, **kwargs):
        from pathlib import Path
        if not kwargs['server_key']:
            raise Exception('Nope!')
        if not kwargs['client_key']:
            raise Exception('Nope!')

        keys_dir = kwargs['keys_path']
        public_keys_dir = keys_dir / 'public_keys'
        secret_keys_dir = keys_dir / 'private_keys'

        if not (keys_dir.exists() and
                public_keys_dir.exists() and
                secret_keys_dir.exists()):
            print("CURVE Server: Certificate missing -- this device needs to be comissioned first!")
            print(public_keys_dir, secret_keys_dir)
            sys.exit(1)

        self.ctx = Kontext()
        self.client = self.ctx.socket(zmq.DEALER)

        client_secret_file = secret_keys_dir / Path(f'{kwargs["client_key"]}.key_secret')
        client_public, client_secret = zmq.auth.load_certificate(client_secret_file)
        self.client.curve_secretkey = client_secret
        self.client.curve_publickey = client_public

        server_public_file = public_keys_dir / Path(f'{kwargs["server_key"]}.key')
        server_public, _ = zmq.auth.load_certificate(server_public_file)
        # The client must know the server's public key to make a CURVE connection.
        self.client.curve_serverkey = server_public
        #woohoo
        retry = 0
        hostname = kwargs['hostname']
        ## FIXME retrying forever is bad practice!
        while True:
            if 'ip' in kwargs:
                break
            retry+=1
            try:
                ip = socket.gethostbyname(hostname)
                kwargs['ip'] = ip
                break
            except Exception as e:
                #print(f'Could not resolve {hostname} - retry {retry}/inf')
                continue
            #if retry == 100:
            #    sys.exit(1)
        # TODO what if the IP changes?!
        # maybe with setsockopt(zmq.heartbeat) ?
        self.client.connect(f'tcp://{kwargs["ip"]}:{kwargs["port"]}')



import zlib
import json
import pickle
import zmq

class GSock(zmq.Socket):
    """A class with some extra serialization methods
    
    """
    def psend(self, msg, compression=False):
        """
        send msg with internal routing as list
        and pickled payload
        """
        # if msg is dict -- return immediately
        # if msg len is 1 -- return immediately
        if isinstance(msg, dict):
            ret = self.send_pyobj(msg)
            return ret
        elif not isinstance(msg, list):
            return True # !!! FIXME
        elif len(msg) <= 1:
            ret = self.send_pyobj(msg)
            return ret

        p = []
        p.extend(msg[:-1])
        if compression:
            comp = zlib.compress(pickle.dumps(msg[-1]))
            p.append(comp)
        else:
            p.append(pickle.dumps(msg[-1]))
        ret = self.send_pyobj(p)
        return ret

    def precv(self, compression=False):
        """
        recv msg with internal routing as list
        and unpickle payload
        """
        msg = self.recv_pyobj()
        if isinstance(msg, dict):
            return msg
        elif not isinstance(msg, list):
            return True # !!! FIXME
        elif len(msg) <= 1:
            return msg

        route = msg[:-1]
        if compression:
            decomp = pickle.loads(zlib.decompress(msg[-1]))
            msg[-1] = decomp
        else:
            msg[-1] = pickle.loads(msg[-1])

        return msg

    def dsend(self, msg, z_route=None, compression=False):
        """
        sending as dealer to dealer (resource worker/client/...)

        emtpy frame for compatibility with req/rep OR zmq internal route
        """
        p = z_route or [b'']

        # if msg is dict return immediately
        if isinstance(msg, dict):
            p.append(pickle.dumps(msg))
            ret = self.send_multipart(p)
            return ret
        elif not isinstance(msg, list):
            return True # !!! FIXME
        elif len(msg) <= 1:
            p.append(pickle.dumps(msg))
            ret = self.send_multipart(p)
            return ret

        # internal app route
        payload = msg[:-1]

        if compression:
            comp = zlib.compress(pickle.dumps(msg[-1]))
            payload.append(comp)
        else:
            payload.append(pickle.dumps(msg[-1]))

        p.append(pickle.dumps(payload))
        ret = self.send_multipart(p)
        return ret

    def drecv(self, compression=False):
        """receiving as dealer from dealer(resource worker)
        compatible with req/rep
        """
        rawmsg = self.recv_multipart()
        route = rawmsg[:-1]
        msg = pickle.loads(rawmsg[-1])
        if isinstance(msg, dict):
            return route, msg
        elif not isinstance(msg, list):
            return True # !!! FIXME
        elif len(msg) <= 1:
            return route, msg

        if compression:
            decomp = pickle.loads(zlib.decompress(msg[-1]))
            msg[-1] = decomp
        else:
             msg[-1] = pickle.loads(msg[-1])
        #print(route, msg)
        return route, msg


class Kontext(zmq.Context):
    _socket_class = GSock



"""
    Copyright (c) 2019 Contributors as noted in the AUTHORS file
    This file is part of ggm, the GILGAMESH core engine in Python.
    ggm is free software; you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License (GPL) as published
    by the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
    As a special exception, the Contributors give you permission to link
    this library with independent modules to produce an executable,
    regardless of the license terms of these independent modules, and to
    copy and distribute the resulting executable under terms of your choice,
    provided that you also meet, for each linked independent module, the
    terms and conditions of the license of that module. An independent
    module is a module which is not derived from or based on this library.
    If you modify this library, you must extend this exception to your
    version of the library.
    ggm is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
    License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import zmq
import random
import json
import time
import asyncio
from zmq.utils.strtypes import asbytes

from ..lib.kernel import GK
from ..lib.kontext import Kontext

from datetime import datetime as dt
from datetime import timedelta as td
from pprint import pprint
from collections import defaultdict

class GData(GK):
    def __init__(self, *args, **kwargs):
        self.loop = kwargs['loop']
        self.context = kwargs['context']
        self.name = kwargs['name']

        self.BUFFER_MAX_AGE = td(seconds=kwargs['buffer_max_age'])
        self.BUFFER_MAX_SIZE = kwargs['buffer_max_size']
        self.VACUUM_INTERVAL = kwargs['vacuum_interval']
        self.buffer_map = defaultdict(dict)

        self.DIN = 'ipc:///tmp/data_in'
        self.PUB = f'tcp://0.0.0.0:{kwargs["port"]}'

        self.FRO = "ipc:///tmp/frontend"
        self.fro = self.context.socket(zmq.DEALER)
        self.fro.connect(self.FRO)

        self.DOUT = 'ipc:///tmp/data_out'
        self.outbound = self.context.socket(zmq.PUSH)
        self.outbound.bind(self.DOUT)

        GK.__init__(self, *args, **kwargs)

        self.loop.create_task(self.relay())
        self.loop.create_task(self.vacuum())

    async def relay(self):
        """doc
        """
        inbound = self.context.socket(zmq.PULL)
        inbound.bind(self.DIN)

        pub = self.context.socket(zmq.PUB)
        pub.bind(self.PUB)
        pub.setsockopt(zmq.CONFLATE, 1)

        while True:
            # zerocopy without pub
            #msg = await inbound.recv(copy=False)
            #await outbound.send(msg, copy=False)

            msg = await inbound.precv()
            #self.glog.debug(f'Got data: {msg}')

            # single points get published
            # data blocks > buffer size get redirected to database immediately
            if len(msg[-1]) == 1:
                pub.send_string(f'{msg[0]} {msg[1]} {json.dumps(msg[-1][0])}')
                # CAVEAT FIXME subscribers need to filter with `["` as prefix
                # no multipart message possible with conflate
                # TESTME maybe it is?!
            elif len(msg[-1]) >= self.BUFFER_MAX_SIZE:
                await self.outbound.psnd(msg)
                continue
            else:
                pass

            db, tb = (msg[0], msg[1])

            # create mapping for new data
            if not db in self.buffer_map:
                self.buffer_map[db] = {}
            if not tb in self.buffer_map[db]:
                self.buffer_map[db][tb] = {}

            # new data needs a data map created
            # if data is empty a new era begins
            if not 'data' in self.buffer_map[db][tb]:
                self.buffer_map[db][tb]['data'] = []
                self.buffer_map[db][tb]['age'] = dt.utcnow()
            elif not len(self.buffer_map[db][tb]['data']):
                self.buffer_map[db][tb]['age'] = dt.utcnow()
            else:
                pass

            # FIXME this should not have to be done here!
            if isinstance(msg[-1], dict):
                msg[-1] = [msg[-1]]

            # change and save tags only on change
            if not 'tags' in self.buffer_map[db][tb]:
                if not 'tags' in msg[-1][-1]:
                    self.buffer_map[db][tb]['tags'] = {}
                else:
                    self.buffer_map[db][tb]['tags'] = msg[-1][-1]['tags']
            elif self.buffer_map[db][tb]['tags'] == msg[-1][-1]['tags']:
                msg[-1][-1]['tags'] = dict.fromkeys(msg[-1][-1]['tags'], None)
            else:
                self.buffer_map[db][tb]['tags'] = msg[-1][-1]['tags']

            #if len(self.buffer_map[db][tb]['data']) >= self.BUFFER_MAX_SIZE:
            #    await self.outbound.psnd([ db, tb, self.buffer_map[db][tb]['data']])
            #    self.buffer_map[db][tb]['data'] = msg[-1]
            #    continue
            #else:
            self.buffer_map[db][tb]['data'].extend(msg[-1])

    async def vacuum(self):
        """doc
        empty buffer map periodically
        """
        while True:
            await asyncio.sleep(self.VACUUM_INTERVAL)
            for db in self.buffer_map.keys():
                for tb in self.buffer_map[db].keys():
                    if ( (dt.utcnow() - self.buffer_map[db][tb]['age']) > self.BUFFER_MAX_AGE
                            or len(self.buffer_map[db][tb]['data']) >= self.BUFFER_MAX_SIZE
                            and len(self.buffer_map[db][tb]['data']) > 0 ):
                        await self.outbound.psnd([db, tb, self.buffer_map[db][tb]['data']])
                        self.buffer_map[db][tb]['data'] = []
                        self.buffer_map[db][tb]['age'] = dt.utcnow()
                    else:
                        pass


def data_kernel_process(kcfg):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    kcfg['context'] = Kontext()
    kcfg['loop'] = loop
    
    data = GData(**kcfg)
    data.start()



"""
    Copyright (c) 2019 Contributors as noted in the AUTHORS file
    This file is part of ggm, the GILGAMESH core engine in Python.
    ggm is free software; you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License (GPL) as published
    by the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
    As a special exception, the Contributors give you permission to link
    this library with independent modules to produce an executable,
    regardless of the license terms of these independent modules, and to
    copy and distribute the resulting executable under terms of your choice,
    provided that you also meet, for each linked independent module, the
    terms and conditions of the license of that module. An independent
    module is a module which is not derived from or based on this library.
    If you modify this library, you must extend this exception to your
    version of the library.
    ggm is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
    License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import sys
import zmq
import time
import json
import asyncio
from zmq.utils.strtypes import asbytes
from ..lib.kontext import Kontext
from ..lib.kernel import GK
from datetime import datetime as dt

from aiohttp import web

from pprint import pprint

class HttpApi(GK):
    def __init__(self, *args, **kwargs):
        self.loop = kwargs['loop']
        self.ctx = kwargs['context']
        self.name = kwargs['name']

        self.api_sock = "ipc:///tmp/frontend"

        GK.__init__(self, *args, **kwargs)

    async def handle_get(self, request):
        # every handler gets it's own connection
        req = self.ctx.socket(zmq.DEALER)
        req.connect(self.api_sock)
        req.setsockopt(zmq.LINGER, 0)

        # chop chop
        call = request.path.split("/")
        params = dict(request.rel_url.query)
        call.pop(0)
        if call[-1] == '':
            call.pop(-1)
        request = ['http-get']
        if params:
            call.append(params)
        request.extend(call)

        # get the requested data
        await req.dsend(request)
        _, reply = await req.drecv()

        # close zmq connection
        req.close()

        return web.json_response(reply)

    async def handle_post(self, request):
        # every handler gets it's own connection
        post = self.ctx.socket(zmq.DEALER)
        post.connect(self.api_sock)
        post.setsockopt(zmq.LINGER, 0)

        # get POST data
        call = request.path.split("/")
        data = await request.json()

        # add timestamp to data if they don't have one
        time_keys = ['time']
        for k in time_keys:
            if k not in data:
                data['time'] = self.get_iso_timestamp()

        # chop chop
        call.pop(0)
        if call[-1] == '':
            call.pop(-1)
        request = ['http-post']
        if data:
            call.append(data)
        request.extend(call)

        # req for data
        await post.dsend(request)
        _, reply = await post.drecv()

        # close zmq connection
        post.close()

        return web.json_response(reply)

def http_api_process(kcfg):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    kcfg['context'] = Kontext()
    kcfg['loop'] = loop
    
    api = HttpApi(**kcfg)

    app = web.Application()
    # catch all requests
    app.router.add_get('/{tail:.*}', api.handle_get)
    app.router.add_post('/{tail:.*}', api.handle_post)
    handler = app.make_handler(
            logger=api.glog,
            access_log=api.glog,
            tcp_keepalive=False,
            keepalive_timeout=3,
            lingering_time=0
            )
    f = loop.create_server(handler, '0.0.0.0', 5555)
    srv = loop.run_until_complete(f)


    api.start()



"""
    Copyright (c) 2019 Contributors as noted in the AUTHORS file
    This file is part of ggm, the GILGAMESH core engine in Python.
    ggm is free software; you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License (GPL) as published
    by the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
    As a special exception, the Contributors give you permission to link
    this library with independent modules to produce an executable,
    regardless of the license terms of these independent modules, and to
    copy and distribute the resulting executable under terms of your choice,
    provided that you also meet, for each linked independent module, the
    terms and conditions of the license of that module. An independent
    module is a module which is not derived from or based on this library.
    If you modify this library, you must extend this exception to your
    version of the library.
    ggm is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
    License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import sys, os
import zmq
import asyncio
import sqlite3
import pickle
import json
from datetime import datetime as dt

from ..lib.kernel import GK
from ..lib.kontext import Kontext

import time
from pprint import pprint

class GJSqlite(GK):
    def __init__(self, *args, **kwargs):
        GK.__init__(self, *args, **kwargs)

        self.DB = "ipc:///tmp/json_out"
        self.data_path = kwargs['data_path']
        try:
            os.makedirs(self.data_path)
        except OSError as e:
            pass

        self.conn = sqlite3.connect(f'{self.data_path}/jsqlite.db', detect_types=sqlite3.PARSE_DECLTYPES)

        self.loop.create_task(self.query_task())

    async def query_task(self):
        """
        task for answering queries
        """
        dbq = self.context.socket(zmq.REP)
        dbq.connect(self.DB)
        cur = self.conn.cursor()
        while True:
            raw = await dbq.precv()
            try:
                sig = raw.pop(0)
                table = raw.pop(0)

                reply = {}

                #print(sig)
                #print(raw)
                if sig == 'measurements':
                    colls = []
                    query = f'SELECT _id FROM "{table}" UNIQE'
                    cur.execute(query)
                    _ids = [i[0] for i in cur.fetchall()]
                    await dbq.psnd(_ids)
                    continue

                elif sig == 'get':
                    _id = raw.pop(0)
                    tmp = raw.pop(0)

                    query = f'SELECT * FROM "{table}" WHERE _id = "{_id}" '

                    if tmp == 'head':
                        query = query+'ORDER BY time DESC LIMIT 1'
                        cur.execute(query)
                        records = cur.fetchone()
                        #await dbq.psnd(pickle.loads(records[-1]))
                        await dbq.psnd(json.loads(records[-1]))
                        continue

                    if tmp == 'tail':
                        query = query+'ORDER BY time ASC LIMIT 1'
                        cur.execute(query)
                        records = cur.fetchone()
                        #await dbq.psnd(pickle.loads(records[-1]))
                        await dbq.psnd(json.loads(records[-1]))
                        continue

                    if tmp == 'all':
                        query = query+'ORDER BY time DESC LIMIT 1000'
                        cur.execute(query)
                        records = cur.fetchall()
                        #await dbq.psnd([pickle.loads(record[-1]) for record in records])
                        await dbq.psnd([json.loads(record[-1]) for record in records])
                        continue

                    # this should never be reached
                    await dbq.psnd({'Error': True, 'Reason': 'Did not get get!' })

                elif sig == 'insert':
                    _id = raw.pop(0)
                    data = raw.pop(0)

                    ret = self.alter_insert(cur, table, _id, data)
                    self.conn.commit()
                    await dbq.psnd(ret)

                elif sig == 'nupsert':
                    _id = raw.pop(0)
                    data = raw.pop(0)

                    cur.execute(f'CREATE TABLE IF NOT EXISTS "{table}" (time TIMESTAMP, _id VARCHAR(128), json BLOB)')
                    query = f'SELECT * FROM "{table}" WHERE _id = "{_id}" ORDER BY time DESC LIMIT 1'
                    cur.execute(query)
                    doc = cur.fetchone()
                    if doc:
                        # rowid? see sqlite docs!
                        query = f'UPDATE "{table}" SET json = ? WHERE rowid IN (SELECT rowid FROM "{table}" WHERE _id = "{_id}" ORDER BY time DESC LIMIT 1)'
                        #doc = pickle.loads(doc[-1])
                        doc = json.loads(doc[-1])
                        #doc.update(data)
                        doc = self.update(doc, data)
                        #cur.execute(query, (pickle.dumps(doc),))
                        cur.execute(query, (json.dumps(doc),))
                    else:
                        resp = self.alter_insert(cur, table, _id, data)

                    self.conn.commit()
                    await dbq.psnd({"Error": False, 'Reason': "NUpsert Success."})

                elif sig == 'delete':
                    _id = raw.pop(0)
                    del_keys = raw

                    try:
                        query = f'SELECT * FROM "{table}" WHERE _id = "{_id}" ORDER BY time DESC LIMIT 1'
                        cur.execute(query)
                        doc = cur.fetchone()
                        doc = json.loads(doc[-1])
                    except Exception as e:
                        await dbq.psend_err(f'fail: this resource does not exist!')
                        continue

                    if not del_keys:
                        query = f'DELETE FROM "{table}" WHERE _id = "{_id}"'
                        cur.execute(query)
                    else:
                        self.delete(doc, del_keys)
                        query = f'UPDATE "{table}" SET json = ? WHERE rowid IN (SELECT rowid FROM "{table}" WHERE _id = "{_id}" ORDER BY time DESC LIMIT 1)'
                        cur.execute(query, (json.dumps(doc),))

                    self.conn.commit()
                    resp = { 'Error': False, 'Reason': f'Delete Success.' }
                    await dbq.psnd(resp)

                else:
                    await dbq.psnd({'Error': True, 'Reason': 'JSQLite Driver does not understand.'})

            except Exception as e:
                self.glog.error(f'Malformed query? -- {e}')
                await dbq.psnd({'Error': True, 'Reason': str(e)})

    def delete(self, d, u):
        k = u.pop(0)
        if not u:
            try:
                del d[k]
            except Exception as e:
                return d
        if isinstance(u, list):
            try:
                self.delete(d[k], u)
            except Exception as e:
                return d
        #return d

    def update(self, d, u):
        for k, v in u.items():
            dv = d.get(k, {})
            if not isinstance(dv, dict):
                d[k] = v
            elif isinstance(v, dict):
                d[k] = self.update(dv, v)
            else:
                d[k] = v
        return d

    def alter_insert(self, cur, table , _id, data, upsert=False):
        """
        """
        try:
            cur.execute(f'CREATE TABLE IF NOT EXISTS "{table}" (time TIMESTAMP, _id VARCHAR(128), json BLOB)')
        except Exception as e:
            self.glog.error(f"could not create table: {e}")
            return { 'Error': True, 'Reason': f'could not create table: {e}' }

        # fkn query
        q = f'INSERT INTO "{table}" (time, _id, json) VALUES (?, ?, ?) '
        try:
            #cur.execute(q, [dt.utcnow().isoformat(sep=' '), _id, pickle.dumps(data)])
            cur.execute(q, [dt.utcnow().isoformat(sep=' '), _id, json.dumps(data)])
            return  { 'Error': False, 'Reason': f'Insert Success.' }
        except Exception as e:
            self.glog.error(f'Could not save data to DB! --- {e}')
            return { 'Error': True, 'Reason': f'psql error: {e}' }

def jsqlite_process(kcfg):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    kcfg['context'] = Kontext()
    kcfg['loop'] = loop
    
    ssqlite = GJSqlite(**kcfg)
    ssqlite.start()



"""
    Copyright (c) 2019 Contributors as noted in the AUTHORS file
    This file is part of ggm, the GILGAMESH core engine in Python.
    ggm is free software; you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License (GPL) as published
    by the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
    As a special exception, the Contributors give you permission to link
    this library with independent modules to produce an executable,
    regardless of the license terms of these independent modules, and to
    copy and distribute the resulting executable under terms of your choice,
    provided that you also meet, for each linked independent module, the
    terms and conditions of the license of that module. An independent
    module is a module which is not derived from or based on this library.
    If you modify this library, you must extend this exception to your
    version of the library.
    ggm is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
    License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
class AioMqttHelper:
    def __init__(self, loop, client):
        self.loop = loop
        self.client = client
        self.client.on_socket_open = self.on_socket_open
        self.client.on_socket_close = self.on_socket_close
        self.client.on_socket_register_write = self.on_socket_register_write
        self.client.on_socket_unregister_write = self.on_socket_unregister_write

    def on_socket_open(self, client, userdata, sock):
        #print("Socket opened")
        pass

        def cb():
            # triggered if socket receives message
            #print("Socket is readable, calling loop_read")
            client.loop_read()

        self.loop.add_reader(sock, cb)
        self.misc = self.loop.create_task(self.misc_loop())

    def on_socket_close(self, client, userdata, sock):
        #print("Socket closed")
        self.loop.remove_reader(sock)
        self.misc.cancel()

    def on_socket_register_write(self, client, userdata, sock):
        #print("Watching socket for writability.")

        def cb():
            #print("Socket is writable, calling loop_write")
            client.loop_write()

        self.loop.add_writer(sock, cb)

    def on_socket_unregister_write(self, client, userdata, sock):
        #print("Stop watching socket for writability.")
        self.loop.remove_writer(sock)

    async def misc_loop(self):
        #print("misc_loop started")
        while self.client.loop_misc() == mqtt.MQTT_ERR_SUCCESS:
            try:
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                break
        #print("misc_loop finished")



"""
    Copyright (c) 2019 Contributors as noted in the AUTHORS file
    This file is part of ggm, the GILGAMESH core engine in Python.
    ggm is free software; you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License (GPL) as published
    by the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
    As a special exception, the Contributors give you permission to link
    this library with independent modules to produce an executable,
    regardless of the license terms of these independent modules, and to
    copy and distribute the resulting executable under terms of your choice,
    provided that you also meet, for each linked independent module, the
    terms and conditions of the license of that module. An independent
    module is a module which is not derived from or based on this library.
    If you modify this library, you must extend this exception to your
    version of the library.
    ggm is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
    License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import sys, os
import zmq
import asyncio
import sqlite3

from ..lib.kernel import GK
from ..lib.kontext import Kontext

import time
from pprint import pprint

class SSqlite(GK):
    def __init__(self, *args, **kwargs):
        GK.__init__(self, *args, **kwargs)

        self.DB = "ipc:///tmp/series_out"
        self.data_path = kwargs['data_path']
        try:
            os.makedirs(self.data_path)
        except OSError as e:
            pass

        # one db per device!
        #self.conn = sqlite3.connect('./ssqlite.db', detect_types=sqlite3.PARSE_DECLTYPES)

        self.DOUT = "ipc:///tmp/data_out"
        self.receiver = self.context.socket(zmq.PULL)
        self.receiver.connect(self.DOUT)
        self.loop.create_task(self.write_data())

        self.loop.create_task(self.query_task())

    async def query_task(self):
        """
        task for answering queries
        """
        dbq = self.context.socket(zmq.REP)
        dbq.connect(self.DB)
        while True:
            raw = await dbq.precv()
            try:
                sig = raw.pop(0)
                dev_id = raw.pop(0)

                #db = sqlite3.connect(f'./{dev_id}.db', detect_types=sqlite3.PARSE_DECLTYPES)
                if sig != 'insert':
                    db = sqlite3.connect(f'file:{self.data_path}/{dev_id}.db?mode=ro', uri=True)
                else:
                    db = sqlite3.connect(f'{self.data_path}/{dev_id}.db')
                cur = db.cursor()
                reply = {}

                #print(sig)
                #print(raw)
                if sig == 'measurements':
                    colls = []
                    query = f'SELECT name FROM sqlite_master WHERE type="table"'
                    cur.execute(query)
                    measurements = [i[0] for i in cur.fetchall()]
                    await dbq.psnd(measurements)
                    cur.close()
                    db.close()
                    continue

                elif sig == 'count':
                    measurement = raw.pop(0)
                    query = f'SELECT count(*) FROM {measurement} '
                    try:
                        tmp = raw.pop(0)
                        if 'start' in tmp.keys():
                            query = query+f"WHERE time >= '{tmp['start']}' "
                            if 'stop' in tmp.keys():
                                query = query+f"AND time < '{tmp['stop']}' "
                    except:
                        pass
                    cur.execute(query)
                    count = [i[0] for i in cur.fetchall()]
                    await dbq.psnd(count)
                    cur.close()
                    db.close()
                    continue

                elif sig == 'get':
                    measurement = raw.pop(0)
                    tmp = raw.pop(0)

                    # get table column names and types for conversion
                    query = f'PRAGMA table_info({measurement})'
                    cur.execute(query)
                    records = cur.fetchall()
                    clist = [i[1] for i in records]
                    tlist = [i[2] for i in records]

                    query = f'SELECT * FROM "{measurement}" '

                    if tmp == 'head':
                        query = query+'ORDER BY time DESC LIMIT 1'

                    elif tmp == 'tail':
                        query = query+'ORDER BY time ASC LIMIT 1'

                    elif tmp == 'all':
                        query = query+'ORDER BY time DESC LIMIT 1000'

                    elif isinstance(tmp, dict):
                        if 'start' in tmp.keys():
                            query = query+f"WHERE time >= '{tmp['start']}' "
                            if 'stop' in tmp.keys():
                                query = query+f"AND time < '{tmp['stop']}' "
                        query = query+ f'ORDER BY time DESC LIMIT 1000'

                    else:
                        await dbq.psnd({'Error': True, 'Reason': 'Did not get get!' })
                        continue

                    records = []
                    cur.execute(query)
                    records = cur.fetchall()
                    records = self.from_db_format(records, clist, tlist)
                    await dbq.psnd(records)
                    cur.close()
                    db.close()
                    continue

                elif sig == 'insert':
                    measurement = raw.pop(0)
                    data = raw.pop(0)

                    if not isinstance(data, list):
                        data = [data]

                    #self.glog.debug(f'inserting {data}')
                    ret = self.alter_insert(cur, measurement, data)
                    db.commit()
                    await dbq.psnd(ret)
                    cur.close()
                    db.close()
                    continue

                elif sig == 'delete':
                    measurement = raw.pop(0)

                    thresh = self.iso_5min_ago()
                    query = f'SELECT deleteable FROM "{measurement}" where time >= "{thresh}" AND deleteable IS NOT NULL ORDER BY time DESC'
                    cur.execute(query)
                    records = cur.fetchall()
                    if records:
                        query = f'DROP TABLE "{measurement}"'
                        cur.execute(query)
                        db.commit()
                        await dbq.psnd({'Error': False, 'Reason': f'Deletion of {measurement} successful.'})
                    else:
                        await dbq.psnd({'Error': True, 'Reason': f'{measurement} is not deleteable.'})

                    cur.close()
                    db.close()
                    continue

                elif sig == 'chunk':
                    tic = time.time()
                    measurement = raw.pop(0)
                    tmp = raw.pop(0)

                    try:
                        offset = tmp['offset']
                        limit = tmp['limit']
                    except Exception as e:
                        await dbq.psnd({'Error': True, 'Reason': str(e)})

                    # get table column names and types for conversion
                    query = f'PRAGMA table_info({measurement})'
                    cur.execute(query)
                    records = cur.fetchall()
                    clist = [i[1] for i in records]
                    tlist = [i[2] for i in records]

                    cols = []
                    if 'cols' in tmp:
                        clist = ['time']
                        tlist = ['TIMESTAMP']
                        for i in records:
                            if i[1] in tmp['cols']:
                                clist.append(i[1])
                                tlist.append(i[2])
                        # FIXME sql-injection-city
                        cols_str = [f'"{col}"' for col in clist[1:]]
                        query = f'SELECT time, {", ".join(cols_str)} FROM "{measurement}" '
                    else:
                        query = f'SELECT * FROM "{measurement}" '

                    if 'start' in tmp.keys():
                        query = query+f"WHERE time > '{tmp['start']}' "
                        if 'stop' in tmp.keys():
                            query = query+f"AND time <= '{tmp['stop']}' "

                    query = query+ f"ORDER BY time DESC LIMIT {limit} OFFSET {offset}"

                    records = []
                    cur.execute(query)
                    records = cur.fetchall()
                    records = self.from_db_format(records, clist, tlist)

                    # get last metadata if necessary
                    zipped = zip(clist, tlist)
                    meta_list = []
                    meta_data = []
                    for i in zipped:
                        if i[1] not in ['TIMESTAMP', 'FLOAT']:
                            meta_list.append(i[0])
                    if records:
                        if not any(list(records[-1]['tags'].values())):
                            if len(meta_list) > 1:
                                query = f'SELECT {", ".join(meta_list)} FROM "{measurement}" WHERE time < "{records[-1]["time"]}" AND COALESCE({", ".join(meta_list)}) IS NOT NULL ORDER BY time DESC LIMIT 1'
                            else:
                                query = f'SELECT {meta_list[0]} FROM "{measurement}" WHERE time < "{records[-1]["time"]}" AND {meta_list[0]} IS NOT NULL ORDER BY time DESC LIMIT 1'
                            cur.execute(query)
                            meta_data = cur.fetchone()
                            records[-1]['tags'] = {k: v for (k, v) in zip(meta_list, meta_data)}

                    # log slow querys
                    toc = time.time()
                    if toc-tic > 2.00:
                        self.glog.info(f'Slow chunk query ({dev_id} {measurement}): {round((toc-tic), 3)} s')

                    await dbq.psnd(records)
                    cur.close()
                    db.close()
                    continue

                else:
                    await dbq.psnd({'Error': True, 'Reason': 'SSQLite Driver does not understand.'})

            except Exception as e:
                self.glog.error(f'Malformed query? --- {e}')
                #self.glog.error(f'Malformed query? --- {e} --- {sig} {dev_id} {measurement}')
                await dbq.psnd({'Error': True, 'Reason': str(e)})

    async def write_data(self):
        while True:
            raw = await self.receiver.precv()

            try:
                dev_id = raw.pop(0)
                measurement = raw.pop(0)
                data = raw.pop(0)
            except Exception as e:
                self.glog.error(f"Error while parsing incoming Data: {e}")
                continue

            if len(data) == 0:
                self.glog.debug(f"incoming data empty")
                continue
            #db = sqlite3.connect(f'./{dev_id}.db', detect_types=sqlite3.PARSE_DECLTYPES)
            db = sqlite3.connect(f'{self.data_path}/{dev_id}.db', timeout=10)
            cur = db.cursor()
            #self.glog.debug(f'Data of len {len(data)} received.')
            self.alter_insert(cur, measurement, data)
            db.commit()
            cur.close()
            db.close()

    def alter_insert(self, cur, measurement, data):
        """
        """
        try:
            cur.execute(f'CREATE TABLE IF NOT EXISTS "{measurement}" (time TIMESTAMP)')
        except Exception as e:
            self.glog.error(f"ssql error: {e}")
            return { 'Error': True, 'Reason': f'ssql error: {e}' }

        #pprint(data)
        clist = ['time']
        clist_str = ['time']
        try:
            cur.execute(f'SELECT * FROM "{measurement}"')
            col_exists = list(map(lambda x: x[0], cur.description))
        except Exception as e:
            self.glog.error(f"ssql error: {e}")
            return { 'Error': True, 'Reason': f'ssql error: {e}' }

        try:
            for col in data[0]['fields']:
                clist.append(col)
                clist_str.append(f'"{col}"')
                if col not in col_exists:
                    cur.execute(f'ALTER TABLE "{measurement}" ADD COLUMN "{col}" FLOAT')
        except Exception as e:
            self.glog.error(f"ssql error: {e}")
            return { 'Error': True, 'Reason': f'ssql error: {e}' }

        try:
            for col in data[0]['tags']:
                clist.append(col)
                clist_str.append(f'"{col}"')
                if col not in col_exists:
                    cur.execute(f'ALTER TABLE "{measurement}" ADD COLUMN "{col}" VARCHAR(128)')
        except Exception as e:
            self.glog.error(f'Could not save data to DB! --- {e} --- {dev_id} {measurement} length: {len(data)}')
            return { 'Error': True, 'Reason': f'ssql error: {e}' }

        # fkn query
        q = f'INSERT INTO "{measurement}" ' + f"({', '.join(clist_str)}) VALUES ({', '.join(['?']*len(clist_str))})"
        data = self.to_db_format(data, clist)
        try:
            cur.executemany(q, data)
            return  { 'Error': False, 'Reason': f'Insert Success.' }
        except Exception as e:
            print(clist)
            pprint(data)
            self.glog.error(f'Could not save data to DB! --- {e}')
            return { 'Error': True, 'Reason': f'ssql error: {e}' }

    def to_db_format(self, data, clist):
        """
        takes list of data dicts and a list with key names
        returns list of data tuples with values from data dict in order of key list
        TODO error checking?
        """
        for i, d in enumerate(data):
            tmp = []
            tmp.append(d['time'])
            for k in clist:
                if k in d['fields'].keys():
                    tmp.append(d['fields'][k])
                elif k in d['tags'].keys():
                    tmp.append(d['tags'][k])
            data[i] = tuple(tmp)
        return data

    def from_db_format(self, data, clist, tlist):
        """
        takes list of data tuples with values from data dict in order of key list
        returns list of data dicts with 'fields' key for float values and 'tags' key for strings
        TODO error checking?
        """
        for i, d in enumerate(data):
            tmp = {}
            tmp['fields'] = {}
            tmp['tags'] = {}
            for j, t in enumerate(tlist):
                if t == 'TIMESTAMP':
                    tmp[clist[j]] = d[j]
                elif t == 'FLOAT':
                    tmp['fields'].update({clist[j]: d[j]})
                else:
                    tmp['tags'].update({clist[j]: d[j]})
            data[i] = tmp
        return data

def ssqlite_process(kcfg):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    kcfg['context'] = Kontext()
    kcfg['loop'] = loop
    
    ssqlite = SSqlite(**kcfg)
    ssqlite.start()



import os
import zmq
import asyncio
from ..lib.kernel import GK
from ..lib.kontext import Kontext

class RPiHealth(GK):
    def __init__(self, *args, **kwargs):
        GK.__init__(self, *args, **kwargs)
        self.UPDATE_INTERVAL = kwargs['interval']

    async def cpu_temp_producer(self):
        """
        get temperature from thermal_zone0 (package temp on raspi)
        get 1min avg load from /proc
        """
        tags=dict()
        tags['temp_unit'] = 'C'
        tags['tags'] = 'monitor,temperature,load'
        measurement_name = 'cpu_temp'
        while True:
            await asyncio.sleep(self.UPDATE_INTERVAL)
            data = float(self.get_cpu_temp())
            tags['load'] = str(self.get_load())
            self.glog.debug("System Temp: {0} C".format(data))
            await self.send_data(self.dev_id, measurement_name, {'temp': data}, tags)

    async def free_mem_producer(self):
        """
        get total memory
        get free memory and calculate free percentage
        """
        tags=dict()
        tags['free_unit'] = 'MB'
        tags['tags'] = 'monitor,free,memory'
        measurement_name = 'free_mem'
        _ , total = self.get_free()
        tags['total'] = str(total)
        while True:
            await asyncio.sleep(self.UPDATE_INTERVAL)
            free, _ = self.get_free()
            tags['percent'] = str(round((free/total)*100, 1))
            self.glog.debug(f'Free Mem: {free}/{total} MB')
            await self.send_data(self.dev_id, measurement_name, {'free': free}, tags)

    def get_cpu_temp(self):
        temp = os.popen('cat /sys/class/thermal/thermal_zone0/temp').readline()
        return float(temp)/1000.0

    def get_load(self):
        load = os.popen('cat /proc/loadavg').readline()
        load_1min = load.split(" ")[0]
        return float(load_1min)

    def get_free(self):
        cmd = '''awk '{ if (/MemAvailable:/) {mem_available=$2}; if (/MemTotal:/) {mem_total=$2};
        if (mem_available && mem_total) {print int(mem_available/1024) " " int(mem_total/1024);exit}}' /proc/meminfo'''
        free, total = os.popen(cmd).readline().split(" ")
        return int(free), int(total)

def rpi_health_process(kcfg):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    kcfg['context'] = Kontext()
    kcfg['loop'] = loop
    
    rpi = RPiHealth(**kcfg)
    loop.create_task(rpi.cpu_temp_producer())
    loop.create_task(rpi.free_mem_producer())
    # gpio control for testing
    # loop.create_task(rpi.control())
    rpi.start()



"""
    Copyright (c) 2019 Contributors as noted in the AUTHORS file
    This file is part of ggm, the GILGAMESH core engine in Python.
    ggm is free software; you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License (GPL) as published
    by the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
    As a special exception, the Contributors give you permission to link
    this library with independent modules to produce an executable,
    regardless of the license terms of these independent modules, and to
    copy and distribute the resulting executable under terms of your choice,
    provided that you also meet, for each linked independent module, the
    terms and conditions of the license of that module. An independent
    module is a module which is not derived from or based on this library.
    If you modify this library, you must extend this exception to your
    version of the library.
    ggm is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
    License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from multiprocessing import Process

import sys
import zmq
import json
import syslog
import logging
from datetime import datetime

import asyncio
from zmq.asyncio import Poller
from zmq.utils.strtypes import asbytes
from zmq.devices import ProcessDevice

from ..lib.data import GData
from ..lib.logging import GLog
from ..lib.heart import GHeartt

from pprint import pprint

class GResourceManager(GHeartt, GLog, GData):
    def __init__(self, kernel_map, name="ResourceManager", *args, **kwargs):
        self.loop = kwargs['loop']
        self.context = kwargs['context']
        self.dev_id = kwargs['device_id']
        self.version = kwargs['version']

        self.kernel_map = kernel_map
        self.name = name

        # whyyyy
        #super().__init__()
        GHeartt.__init__(self, *args, **kwargs)
        GData.__init__(self, *args, **kwargs)
        GLog.__init__(self, *args, **kwargs)

        # 'frontend' <-> 'backend' proxy (polled)
        self.FRO = "ipc:///tmp/frontend"
        self.BAC = "ipc:///tmp/backend"
        #self.fro = self.context.socket(zmq.ROUTER)
        #self.fro.bind(self.FRO)
        #self.bac = self.context.socket(zmq.DEALER)
        #self.bac.bind(self.BAC)
        ## Initialize poll set
        #self.poller = Poller()
        #self.poller.register(self.fro, zmq.POLLIN)
        #self.poller.register(self.bac, zmq.POLLIN)
        # FIXME/TESTME
        # split relay off of resource manager as process device?
        # or own kernel?
        # 'frontend' <-> 'backend' proxy process device
        devx = ProcessDevice(zmq.QUEUE, zmq.ROUTER, zmq.DEALER)
        devx.bind_in(self.FRO)
        devx.bind_out(self.BAC)
        devx.start()

        # resource_worker <-> database proxy for storage workers
        for res_type in kwargs['resource_types']:
            sdev = ProcessDevice(zmq.QUEUE, zmq.ROUTER, zmq.DEALER)
            sdev.bind_in(f'ipc:///tmp/{res_type}_in')
            sdev.bind_out(f'ipc:///tmp/{res_type}_out')
            sdev.start()

    async def resource_task(self):
        """
        pipeline for resource requests to resource workers
        FIXME implement inter broker routing here?
        FIXME tight now a process device is used. this is dead code?
        """
        while True:
            socks = dict(await self.poller.poll())

            if socks.get(self.fro) == zmq.POLLIN:
                message = await self.fro.recv_multipart(copy=False)
                #self.glog.debug("frontend in: {0}".format(message))
                await self.bac.send_multipart(message, copy=False)

            if socks.get(self.bac) == zmq.POLLIN:
                message = await self.bac.recv_multipart(copy=False)
                #self.glog.debug("frontend out: {0}".format(message))
                await self.fro.send_multipart(message, copy=False)

    async def start_kernels(self):
        # start delayed wait for logger etc...
        await asyncio.sleep(1)
        for n, kernel in enumerate(self.kernel_map):
            self.init_kernel(kernel, n)
            self.config_kernel(kernel, n)
        return

    def init_kernel(self, kernel, n):
        self.kernel_map[n]['proc'] = Process(target=kernel['function'], args=(kernel['config'],))
        self.kernel_map[n]['proc'].start()
        self.glog.info(f'Starting Kernel {kernel["name"]} with PID {self.kernel_map[n]["proc"].pid}')

    # TODO
    # MAYBE???
    def config_kernel(self, kernel, n):
        """
        set log level
        set logging True/False
        pause kernel?
        """
        pass

    async def logger_task(self):
        """TODO: Fix logger formatting and message handling!!!!"""
        """TODO: filter logging for kernels (self.kernel_map) OR control kernel logging behavior!!!!"""
        s = self.context.socket(zmq.SUB)
        s.bind(self.LOG_DEST)
        s.subscribe(b'')
        while True:
            msg = await s.recv_multipart()
            #print(msg)
            level, message = msg
            message = message.decode('ascii')
            if message.endswith('\n'):
                message = message[:-1]
            log = getattr(logging, level.lower().decode('ascii'))
            syslog.syslog(message)

    def start(self):
        # try me!
        #self.loop.create_task(self.resource_task())
        self.loop.create_task(self.logger_task())
        self.loop.create_task(self.heartbeater())
        self.loop.create_task(self.handle_pong())
        self.loop.create_task(self.start_kernels())

    def terminate(self):
        return



"""
    Copyright (c) 2019 Contributors as noted in the AUTHORS file
    This file is part of ggm, the GILGAMESH core engine in Python.
    ggm is free software; you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License (GPL) as published
    by the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
    As a special exception, the Contributors give you permission to link
    this library with independent modules to produce an executable,
    regardless of the license terms of these independent modules, and to
    copy and distribute the resulting executable under terms of your choice,
    provided that you also meet, for each linked independent module, the
    terms and conditions of the license of that module. An independent
    module is a module which is not derived from or based on this library.
    If you modify this library, you must extend this exception to your
    version of the library.
    ggm is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
    License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import sys
import zmq
import asyncio

import psycopg2
import psycopg2.extras
from psycopg2 import sql

from ..lib.kernel import GK
from ..lib.kontext import Kontext

import time
import json
from pprint import pprint
from datetime import datetime as dt

class KTimescale(GK):
    def __init__(self, *args, **kwargs):
        GK.__init__(self, *args, **kwargs)

        self.DB = "ipc:///tmp/series_out"

        # FIXME hard coded? better for secrets?
        with open(f'/etc/gilgamesh/timescaledb.secret', 'r') as raw:
            dbcfg = json.load(raw)
            dbname = dbcfg['dbname']
            dbuser = dbcfg['dbuser']
            dbpw = dbcfg['dbpw']
            dbhost = dbcfg['dbhost']
            del dbcfg

        self.dsn = f'dbname={dbname} user={dbuser} password={dbpw} host={dbhost}'
        #self.conn = psycopg2.connect(dsn)
        #self.conn.set_isolation_level(0) # autocommit
        #tmst_as_string = psycopg2.extensions.new_type(psycopg2.extensions.TIME.values, 'TIME', psycopg2.STRING)
        #psycopg2.extensions.register_type(tmst_as_string)

        # we are a time series adapter!
        self.DOUT = "ipc:///tmp/data_out"
        self.receiver = self.context.socket(zmq.PULL)
        self.receiver.connect(self.DOUT)
        self.loop.create_task(self.write_data())

        self.loop.create_task(self.query_task())

    async def query_task(self):
        """
        task for answering timescaledb queries
        """
        dbq = self.context.socket(zmq.REP)
        dbq.connect(self.DB)
        conn = psycopg2.connect(self.dsn)
        conn.set_session(autocommit=True, readonly=True)
        cur = conn.cursor()
        while True:
            raw = await dbq.precv()
            try:
                sig = raw.pop(0)
                dev_id = raw.pop(0)

                reply = {}

                #print(sig)
                #pprint(raw)
                if sig == 'measurements':
                    colls = []
                    query = sql.SQL("SELECT table_name FROM information_schema.tables WHERE table_schema = {0}").format(sql.Literal(dev_id))
                    cur.execute(query)
                    measurements = [i[0] for i in cur.fetchall()]
                    await dbq.psnd(measurements)
                    continue

                elif sig == 'count':
                    measurement = raw.pop(0)
                    query = sql.SQL("SELECT COUNT(*) FROM {0}.{1} ").format(*tuple(map(sql.Identifier, [dev_id, measurement])))

                    try:
                        tmp = raw.pop(0)
                        if 'start' in tmp.keys():
                            query = query + sql.SQL(f"WHERE time >= '{tmp['start']}' ")
                            if 'stop' in tmp.keys():
                                query = query + sql.SQL(f"AND time < '{tmp['stop']}' ")
                    except:
                        pass

                    cur.execute(query)
                    count = [i[0] for i in cur.fetchall()]
                    await dbq.psnd(count)
                    continue

                elif sig == 'get':
                    measurement = raw.pop(0)
                    tmp = raw.pop(0)

                    query = sql.SQL("SELECT * FROM {0}.{1} ").format(*tuple(map(sql.Identifier, [dev_id, measurement])))

                    if tmp == 'head':
                        query = query + sql.SQL('ORDER BY time DESC LIMIT 1')
                        cur.execute(query)
                        records = cur.fetchall()
                        clist = [i[0] for i in cur.description]
                        records = self.from_db_format(records, clist)
                        await dbq.psnd(records)
                        continue

                    if tmp == 'tail':
                        query = query + sql.SQL('ORDER BY time ASC LIMIT 1')
                        cur.execute(query)
                        records = cur.fetchall()
                        clist = [i[0] for i in cur.description]
                        records = self.from_db_format(records, clist)
                        await dbq.psnd(records)
                        continue

                    if tmp == 'all':
                        query = query + sql.SQL('ORDER BY time DESC LIMIT 1000')
                        cur.execute(query)
                        records = cur.fetchall()
                        clist = [i[0] for i in cur.description]
                        records = self.from_db_format(records, clist)
                        await dbq.psnd(records)
                        continue

                    if isinstance(tmp, dict):
                        if 'start' in tmp.keys():
                            query = query + sql.SQL(f"WHERE time >= '{tmp['start']}' ")
                            if 'stop' in tmp.keys():
                                query = query + sql.SQL(f"AND time < '{tmp['stop']}' ")

                        query = sql.Composed(query, 'ORDER BY time DESC LIMIT 1000')
                        records = []

                        cur.execute(query)
                        records = cur.fetchall()
                        clist = [i[0] for i in cur.description]
                        records = self.from_db_format(records, clist)
                        await dbq.psnd(records)
                        continue
                    # this should never be reached
                    await dbq.psnd({'Error': True, 'Reason': 'Did not get get!' })

                elif sig == 'insert':
                    measurement = raw.pop(0)
                    data = raw.pop(0)

                    # should not be needed !!
                    # remove when sure!
                    if not isinstance(data, list):
                        data = [data]

                    ret = self.alter_insert(cur, dev_id, measurement, data)
                    await dbq.psnd(ret)
                    continue

                elif sig == 'delete':
                    measurement = raw.pop(0)

                    thresh = self.iso_5min_ago().isoformat(sep=' ')
                    query = sql.SQL("SELECT EXISTS(SELECT deleteable FROM {0}.{1} where time >= {2} AND deleteable IS NOT NULL ORDER BY time DESC)").format(
                            sql.Identifier(dev_id), sql.Identifier(measurement), sql.Literal(thresh))
                    cur.execute(query)
                    records = cur.fetchall()
                    if records[0][0]:
                        query = sql.SQL("DROP TABLE {0}.{1} CASCADE").format(
                                sql.Identifier(dev_id), sql.Identifier(measurement))
                        cur.execute(query)
                        await dbq.psnd({'Error': False, 'Reason': f'Deletion of {measurement} successful.'})
                    else:
                        await dbq.psnd({'Error': True, 'Reason': f'{measurement} is not deleteable.'})

                    continue

                elif sig == 'chunk':
                    tic = time.time()
                    measurement = raw.pop(0)
                    tmp = raw.pop(0)
                    try:
                        offset = tmp['offset']
                        limit = tmp['limit']
                    except Exception as e:
                        await dbq.psnd({'Error': True, 'Reason': str(e)})

                    query = sql.SQL("SELECT column_name, data_type FROM information_schema.columns WHERE table_schema = {0} AND table_name = {1}").format(sql.Literal(dev_id),sql.Literal(measurement))
                    cur.execute(query)
                    records = cur.fetchall()
                    meta_list = [x[0] for x in filter(lambda x: x[1] == 'character varying', records)]

                    clist = []
                    if 'cols' in tmp:
                        for i in records:
                            if i[0] == 'time':
                                clist.append(i[0])
                            elif i[0] in tmp['cols']:
                                clist.append(i[0])
                        # FIXME sql-injection-city
                        query = sql.SQL("SELECT {2} FROM {0}.{1} ").format(
                                *tuple(map(sql.Identifier, [dev_id, measurement])),
                                sql.SQL(', ').join(map(sql.Identifier, clist)))
                    else:
                        clist = [n[0] for n in records]
                        query = sql.SQL("SELECT * FROM {0}.{1} ").format(
                                sql.Identifier(dev_id), sql.Identifier(measurement))

                    # FIXME sql-injection-city
                    if 'start' in tmp.keys():
                        query = query + sql.SQL(f"WHERE time > '{tmp['start']}' ")
                        if 'stop' in tmp.keys():
                            query = query + sql.SQL(f"AND time <= '{tmp['stop']}' ")

                    query = query + sql.SQL(f"ORDER BY time DESC LIMIT {limit} OFFSET {offset}")

                    records = []
                    cur.execute(query)
                    records = cur.fetchall()
                    records = self.from_db_format(records, clist)

                    # get last metadata if necessary
                    if records:
                        if not any(list(records[-1]['tags'].values())):
                            query = sql.SQL("SELECT {0} FROM {1}.{2} WHERE time < {3} AND COALESCE({0}) IS NOT NULL ORDER BY time DESC LIMIT 1").format(
                                    sql.SQL(', ').join(map(sql.Identifier, meta_list)),
                                    *tuple(map(sql.Identifier, [dev_id, measurement])),
                                    sql.Literal(records[-1]['time']))
                            cur.execute(query)
                            meta_data = cur.fetchone()
                            records[-1]['tags'] = {k: v for (k, v) in zip(meta_list, meta_data)}

                    # log slow querys
                    toc = time.time()
                    if toc-tic > 2.00:
                        self.glog.info(f'Slow chunk query ({dev_id} {measurement}): {round((toc-tic), 3)} s')

                    await dbq.psnd(records)
                    continue

                else:
                    await dbq.psnd({'Error': True, 'Reason': 'TimescaleDB Driver does not understand.'})

            except Exception as e:
                self.glog.error(f'Malformed query? -- {e}')
                await dbq.psnd({'Error': True, 'Reason': str(e)})

    async def write_data(self):
        conn = psycopg2.connect(self.dsn)
        conn.set_session(autocommit=True) # autocommit
        conn.set_isolation_level(0) # autocommit
        cur = conn.cursor()
        while True:
            raw = await self.receiver.precv()

            try:
                dev_id = raw.pop(0)
                measurement = raw.pop(0)
                data = raw.pop(0)
            except Exception as e:
                self.glog.error(f"Error while parsing incoming Data: {e}")
                continue
            if len(data) == 0:
                self.glog.debug(f"incoming data empty")
                continue
            #self.glog.debug(f'Data of len {len(data)} received.')
            self.alter_insert(cur, dev_id, measurement, data)

    def alter_insert(self, cur, dev_id, measurement, data):
        """
        """
        try:
            cur.execute(sql.SQL("CREATE SCHEMA IF NOT EXISTS {0}").format(sql.Identifier(dev_id)))
            cur.execute(sql.SQL("CREATE TABLE IF NOT EXISTS {0}.{1} (time TIMESTAMP)").format(
                *tuple(map(sql.Identifier, [dev_id, measurement]))))
            cur.execute(sql.SQL("SELECT create_hypertable('{0}.{1}', 'time', chunk_time_interval => INTERVAL '1 day', if_not_exists => TRUE)").format(
                *tuple(map(sql.Identifier, [dev_id, measurement]))))
        except Exception as e:
            self.glog.error(f"psql error: {e}")
            return { 'Error': True, 'Reason': f'psql error: {e}' }

        clist = ['time']
        try:
            keys = ['fields', 'tags']
            for key in keys:
                if key not in data[0]:
                    data[0][key] = {}
            for col in data[0]['fields']:
                clist.append(col)
                cur.execute(sql.SQL('ALTER TABLE {0}.{1} ADD COLUMN IF NOT EXISTS {2} FLOAT DEFAULT NULL').format(
                    *tuple(map(sql.Identifier, [dev_id, measurement, col]))))
            for col in data[0]['tags']:
                clist.append(col)
                cur.execute(sql.SQL('ALTER TABLE {0}.{1} ADD COLUMN IF NOT EXISTS {2} VARCHAR(128) DEFAULT NULL').format(
                    *tuple(map(sql.Identifier, [dev_id, measurement, col]))))
        except Exception as e:
            self.glog.error(f'Could not save data to DB! --- {e} --- {dev_id} {measurement} length: {len(data)}')
            return { 'Error': True, 'Reason': f'psql error: {e}' }

        # fkn query
        #q = sql.SQL("INSERT INTO {0}.{1} ({2}) VALUES ({3})").format(
        #        *tuple(map(sql.Identifier, [dev_id, measurement])),
        #        sql.SQL(', ').join(map(sql.Identifier, clist)),
        #        sql.SQL(', ').join(sql.Placeholder() * len(clist)))
        q = sql.SQL("INSERT INTO {0}.{1} ({2}) VALUES %s").format(
                *tuple(map(sql.Identifier, [dev_id, measurement])),
                sql.SQL(', ').join(map(sql.Identifier, clist)))
        data = self.to_db_format(data, clist)
        try:
            #psycopg2.extras.execute_batch(cur, q, data)
            psycopg2.extras.execute_values(cur, q, data)
            return  { 'Error': False, 'Reason': f'Insert Success.' }
        except Exception as e:
            self.glog.error(f'Could not save data to DB! --- {e} --- {dev_id} {measurement} length: {len(data)}')
            return { 'Error': True, 'Reason': f'psql error: {e}' }

    def to_db_format(self, data, clist):
        """
        takes list of data dicts and a list with key names
        returns list of data tuples with values from data dict in order of key list
        TODO error checking?
        """
        keys = ['fields', 'tags']
        for i, d in enumerate(data):
            tmp = []
            tmp.append(d['time'])
            for key in keys:
                if key not in d:
                    d[key] = {}
            for k in clist:
                if k in d['fields'].keys():
                    tmp.append(d['fields'][k])
                elif k in d['tags'].keys():
                    tmp.append(d['tags'][k])
            data[i] = tuple(tmp)
        return data

    def from_db_format(self, data, clist):
        """
        takes list of data tuples with values from data dict in order of key list
        returns list of data dicts with 'fields' key for float values and 'tags' key for strings
        TODO error checking?
        """
        for i, d in enumerate(data):
            tmp = {}
            tmp['fields'] = {}
            tmp['tags'] = {}
            for j, k in enumerate(clist):
                if k == 'time':
                    tmp[k] = d[j].isoformat(sep=' ')
                elif type(d[j]) in [int, float]:
                    tmp['fields'].update({k: d[j]})
                elif type(d[j]) is str:
                    tmp['tags'].update({k: d[j]})
            data[i] = tmp
        return data

def timescale_process(kcfg):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    kcfg['context'] = Kontext()
    kcfg['loop'] = loop
    
    timescale = KTimescale(**kcfg)
    timescale.start()



import zmq
import json
import asyncio
from ..lib.kernel import GK
from ..lib.aiomqtthelper import AioMqttHelper
from ..lib.kontext import Kontext

import paho.mqtt.client as mqtt
import socket

from datetime import datetime as dt  # this
from datetime import timedelta as td # is
from datetime import timezone as tz  # madness

class GMqttBridge(GK, AioMqttHelper):
    def __init__(self, *args, **kwargs):
        GK.__init__(self, *args, **kwargs)

    def on_connect(self, client, userdata, flags, rc):
        print("Connecting")
    
    def on_message(self, client, userdata, message):
        #if not self.got_message:
        #    self.glog.debug("Got unexpected message: {}".format(msg.decode()))
        #else:
        #    self.got_message.set_result(json.loads(message.payload.decode()))
        self.got_message.set_result(message)

    def on_disconnect(self, client, userdata, rc):
        self.disconnected.set_result(rc)

    async def paho_loop(self):
        with open(f'/etc/gilgamesh/mqtt.secret', 'r') as raw:
            cfg = json.load(raw)

        self.disconnected = self.loop.create_future()
        self.got_message = None

        self.client = mqtt.Client()
        self.client.username_pw_set(cfg['username'], cfg['password'])
        self.client.on_connect = self.on_connect
        self.client.on_message = self.on_message
        self.client.on_disconnect = self.on_disconnect

        aioh = AioMqttHelper(self.loop, self.client)

        # fixme -> get all from config
        self.client.connect(cfg['hostname'], 1883, 60)
        self.client.socket().setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 2048)
        self.client.subscribe(f'gilgamesh/data_in/#')

        del cfg

        while True:
            await asyncio.sleep(1)
            self.got_message = self.loop.create_future()
            msg = await self.got_message
            self.got_message = None

            try:
                topic = msg.topic.split('/')
                topic.pop(0)
                dev_id = topic.pop(0)
                coll = topic.pop(0)
                payload = msg.payload.decode()
            except Exception as e:
                self.glog.error(f'got a bad mqtt message: {e}')

            print(dev_id, coll)
            print(payload)
            #tags = None
            #data = {}
            #tmst = dt.timestamp(dt.utcfromtimestamp(float(msg["time"]) + float(msg["msec"])/1000))
            #self.send_data(dev_id, res_name, data, unit, tags, tmst)

def mqtt_bridge_process(kcfg):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    kcfg['context'] = Kontext()
    kcfg['loop'] = loop

    bri = GMqttBridge(**kcfg)
    loop.create_task(bri.paho_loop())
    bri.start()



"""
    Copyright (c) 2019 Contributors as noted in the AUTHORS file
    This file is part of ggm, the GILGAMESH core engine in Python.
    ggm is free software; you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License (GPL) as published
    by the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
    As a special exception, the Contributors give you permission to link
    this library with independent modules to produce an executable,
    regardless of the license terms of these independent modules, and to
    copy and distribute the resulting executable under terms of your choice,
    provided that you also meet, for each linked independent module, the
    terms and conditions of the license of that module. An independent
    module is a module which is not derived from or based on this library.
    If you modify this library, you must extend this exception to your
    version of the library.
    ggm is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
    License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import zmq
import json
import asyncio
from zmq.utils.strtypes import asbytes
from datetime import datetime

from .logging import GLog
from .heart import GHeartt

from datetime import datetime as dt
from datetime import timedelta as td
from datetime import timezone as tz

class GData(object):
    """
        TODO: write doc
    """
    def __init__(self, *args, **kwargs):

        self.DIN = "ipc:///tmp/data_in"
        #self.DOUT = "ipc:///tmp/data_out"

        self.sender = self.context.socket(zmq.PUSH)
        self.sender.connect(self.DIN)

        #self.receiver = self.context.socket(zmq.SUB)
        #self.receiver.connect(self.DOUT)

    async def send_data(self, dev_id, res_name, value, tags, tmst=None):
        """
        timestamp MUST be python datetime object
        """
        tmst = tmst or self.get_iso_timestamp()

        # internal 'route'
        msg = [dev_id, res_name]
        # always list of dicts so we can treat batch data the same 
        data = [
            {
                'time': tmst,
                'fields': value,
                'tags': tags
            }
        ]

        msg.append(data)
        #self.glog.debug(data)
        await self.sender.psnd(msg)

    async def send_batch(self, msg):
        await self.sender.psnd(msg)

    def iso_5min_ago(self):
        return dt.utcnow() - td(minutes=5)

    def get_iso_timestamp(self):
        return dt.utcnow().isoformat(sep=' ') # sep needed for better sqlite compatibility

    def to_timestamp(self, dati):
        return dt.timestamp(dati.replace(tzinfo=tz.utc)) # whyyyyy

    def to_datetime(self, tmst):
        return dt.utcfromtimestamp(tmst) # oh whyyy

    def convert_to_timestamp(self, doc): # whyyy
        """
        recursive function
        converts values of specific keys in a nested dict
        from python datetime to timestamp (utc)
        horrible efficiency?
        """
        timestamp_keys = ['time', 'seen', 'head', 'tail', 'start', 'stop']

        if isinstance(doc, list):
            for thing in doc:
                self.convert_to_timestamp(thing)
        elif isinstance(doc, dict):
            for key in doc.keys():
                if isinstance(doc[key], dict) or isinstance(doc[key], list):
                    self.convert_to_timestamp(doc[key])
                elif key in timestamp_keys:
                    try:
                        doc[key] = self.to_timestamp(doc[key])
                    except Exception as e:
                        self.glog.error("Failed converting timestamp to datetime: {0}".format(e))
        return doc

    def convert_to_datetime(self, doc): # whyyyyyyyy
        """
        recursive function
        converts values of specific keys in a nested dict
        from utc timestamps as floats to python datetime object
        horrible efficiency?
        """
        timestamp_keys = ['time', 'seen', 'head', 'tail', 'start', 'stop']

        if isinstance(doc, list):
            for thing in doc:
                self.convert_to_datetime(thing)
        elif isinstance(doc, dict):
            for key in doc.keys():
                if isinstance(doc[key], dict) or isinstance(doc[key], list):
                    self.convert_to_datetime(doc[key])
                elif key in timestamp_keys:
                    try:
                        doc[key] = self.to_datetime(doc[key])
                    except Exception as e:
                        self.glog.error("Failed converting timestamp to datetime: {0}".format(e))
        return doc



"""
    Copyright (c) 2019 Contributors as noted in the AUTHORS file
    This file is part of ggm, the GILGAMESH core engine in Python.
    ggm is free software; you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License (GPL) as published
    by the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
    As a special exception, the Contributors give you permission to link
    this library with independent modules to produce an executable,
    regardless of the license terms of these independent modules, and to
    copy and distribute the resulting executable under terms of your choice,
    provided that you also meet, for each linked independent module, the
    terms and conditions of the license of that module. An independent
    module is a module which is not derived from or based on this library.
    If you modify this library, you must extend this exception to your
    version of the library.
    ggm is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
    License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import zmq
import asyncio

from ..lib.kernel import GK
from ..lib.kontext import Kontext

class KResourceWorker(GK):
    """
    provides a basic REST-like API to resources

    """
    def __init__(self, *args, **kwargs):
        GK.__init__(self, *args, **kwargs)

        self.storage_cfg = kwargs['resource_types']

        self.BAC = 'ipc:///tmp/backend'
        self.bac = self.context.socket(zmq.DEALER)
        self.bac.connect(self.BAC)

        for res_type in self.storage_cfg:
            if res_type == 'json':
                self.json_DB = f'ipc:///tmp/json_in'
                self.json_db = self.context.socket(zmq.REQ)
                self.json_db.connect(self.json_DB)
            elif res_type == 'series':
                self.series_DB = f'ipc:///tmp/series_in'
                self.series_db = self.context.socket(zmq.REQ)
                self.series_db.connect(self.series_DB)
            else:
                self.glog.error('No known storage adapter configured!')


        # tricky!! (context/socket wise)
        #count = kwargs['count']
        #for i in range(count):
        #    self.loop.create_task(self.work(i))
        self.loop.create_task(self.work())

    async def get_all(self, resource_sock, dbs, colls, cmd):
        reply = {}
        for i, db in enumerate(dbs):
            reply[db] = {}
            for coll in colls[i]:
                query = ['get', db, coll, cmd]
                await resource_sock.psnd(query)
                reply[db][coll] = await resource_sock.precv()
        return reply

    async def work(self, i=0):
        """
        this is the glue between
        application logic and backend logic
        it aims to provide peaceful coexistence of multiple protocols

        TODO
        *zero copy* for payload (last frame in multipart) only would be nice
        """
        me = i
        while True:
            # 'filter' flags
            http_get = False
            http_post = False

            z_route, raw = await self.bac.drecv()
            #print(f'RW: {z_route} {raw}')
            #self.glog.info(f'RW: {z_route} {raw}')
            try:
                # unwrap present
                sig = raw.pop(0)
                sig_typ = sig
            except Exception as e:
                self.glog.error(f'RW({me}) received strange signal -- {0}'.format(e))
                await self.bac.dsend_err('fuck you too!', z_route)
                continue

            #print(sig)
            """
            FIXME: This should be protocol code! (e.g. flatbuffers?)
            FIXME: Should this be protocol code? or zmq router/dealer identity?
            """
            #
            ## sieve http requests
            #
            if sig == 'http-get':
                try:
                    sig = raw.pop(0)
                    sig_typ = sig
                except:
                    await self.bac.dsend_err('', z_route)
                    continue
                if sig not in ['json', 'series']:
                    await self.bac.dsend_err('http-get signal error.', z_route)
                    continue
                http_get = True
            elif sig == "http-post":
                try:
                    sig = raw.pop(0)
                    sig_typ = sig
                except:
                    await self.bac.dsend_err('', z_route)
                    continue
                if sig not in ['json', 'series']:
                    await self.bac.dsend_err('http-post signal error.', z_route)
                    continue
                http_post = True
            else:
                pass

            #
            ## sieve/validate
            #
            if sig == "greetings":
                await self.bac.dsend(['earthlings'], z_route)
                continue
            elif sig == "info":
                await self.bac.dsend({'dev_id': self.dev_id, 'version': self.version}, z_route)
                continue
            #elif sig == "cache":
            #    try:
            #        sig = raw[0]
            #    except:
            #        await self.bac.dsend_err('', z_route)
            #        continue
            #    if sig not in ['get', 'insert']:
            #        await self.bac.dsend_err('cache signal error.', z_route)
            #        continue
            #    resource_sock = self.cache
            elif sig == "series":
                try:
                    sig = raw[0]
                except:
                    await self.bac.dsend_err('', z_route)
                    continue
                if http_get:
                    if sig not in ['get']:
                        await self.bac.dsend_err('http-get signal error.', z_route)
                        continue
                if http_post:
                    if sig not in ['insert']:
                        await self.bac.dsend_err('http-post signal error.', z_route)
                        continue
                if sig not in ['measurements', 'count', 'get', 'upload', 'insert', 'delete', 'chunk', 'inventory']:
                    await self.bac.dsend_err('series signal error.', z_route)
                    continue
                resource_sock = self.series_db
            elif sig == "json":
                try:
                    sig = raw[0]
                except:
                    await self.bac.dsend_err('', z_route)
                    continue
                if http_get:
                    if sig not in ['get']:
                        await self.bac.dsend_err('http-get signal error.', z_route)
                        continue
                if http_post:
                    if sig not in ['insert', 'nupsert']:
                        await self.bac.dsend_err('http-post signal error.', z_route)
                        continue
                if sig not in ['get', 'insert', 'nupsert', 'delete']:
                    await self.bac.dsend_err('json signal error.', z_route)
                    continue
                resource_sock = self.json_db
            else:
                await self.bac.dsend_err('signal error.', z_route)
                continue


            # redirect directly if nostorage is being used
            # FIXME loop over resource types -> if storage_cfg[resource_type] == 'nostorage'
            if ((self.storage_cfg["series"] == 'nostorage' and sig_typ == 'series')
                or (self.storage_cfg["json"] == 'nostorage' and sig_typ == 'json')):
                raw.insert(0, sig_typ)
                await resource_sock.psnd(raw)
                reply = await resource_sock.precv()
                await self.bac.dsend(reply, z_route)
                continue
            #
            ## operations
            #
            #print(raw)
            if sig == "get":
                try:
                    cmd = raw.pop(0)
                    tmp = raw.pop(0)
                except Exception as e:
                    self.glog.error(f'Error while parsing get: {e}')
                    await self.bac.dsend_err('No database specified.', z_route)
                    continue

                if tmp == 'all':
                    dbs = [self.dev_id]
                else:
                    dbs = [tmp]

                try:
                    tmp = raw.pop(0)
                except Exception as e:
                    self.glog.error(f'Error while parsing get: {e}')
                    await self.bac.dsend_err('No collection specified.', z_route)
                    continue

                if tmp == 'all':
                    colls = []
                    for db in dbs:
                        query = ['measurements', db]
                        await resource_sock.psnd(query)
                        colls.append(await resource_sock.precv())
                else:
                    colls = [[tmp]]

                try:
                    cmds = raw.pop(0)
                except Exception as e:
                    self.glog.error(f"Error while parsing get: {e}")
                    await self.bac.dsend_err('No action specified.', z_route)
                    continue

                reply = {}
                reply = await self.get_all(resource_sock, dbs, colls, cmds)

                await self.bac.dsend(reply, z_route)
                continue

            elif sig == 'insert':
                """
                insert document in collection
                """
                try:
                    cmd = raw.pop(0)
                    db = raw.pop(0)
                    coll = raw.pop(0)
                    doc = raw.pop(0)
                except Exception as e:
                    self.glog.error(f"Error while parsing insert: {e}")
                    await self.bac.dsend_err('Error parsing insert query.', z_route)
                    continue

                query = [cmd, db, coll, doc]

                if sig_typ == 'series':
                    query.pop(0)
                    await self.send_batch(query)
                    await self.bac.dsend({'Error': False, 'Reason': 'all is good'}, z_route)
                    continue
                else:
                    await resource_sock.psnd(query)
                    reply = await resource_sock.precv()
                    await self.bac.dsend(reply, z_route)
                    continue

            elif sig == 'nupsert':
                """
                update newest document in collection or create one and insert
                """
                try:
                    cmd = raw.pop(0)
                    db = raw.pop(0)
                    coll = raw.pop(0)
                    doc = raw.pop(0)
                except Exception as e:
                    self.glog.error(f"Error while parsing nupsert: {e}")
                    await self.bac.dsend_err('Error parsing nupsert query.', z_route)
                    continue

                query = [cmd, db, coll, doc]

                await resource_sock.psnd(query)
                reply = await resource_sock.precv()

                await self.bac.dsend(reply, z_route)
                continue

            elif sig == 'delete':
                """
                delete
                only tables are deleteable (no databases)
                tables need a column called deleteable with a value within the last 5min
                """
                try:
                    cmd = raw.pop(0)
                    db = raw.pop(0)
                    coll = raw.pop(0)
                    query = [cmd, db, coll]
                    # jsqlite demands more 'levels'
                    if isinstance(raw, list) and len(raw) > 0:
                        query.extend(raw)
                except Exception as e:
                    self.glog.error(f"Error while parsing delete: {e}")
                    await self.bac.dsend_err('Error parsing delete query.', z_route)
                    continue

                await resource_sock.psnd(query)
                reply = await resource_sock.precv()

                await self.bac.dsend(reply, z_route)
                continue

            elif sig == "upload":
                raw.pop(0)
                if len(raw) > 0: 
                    await self.send_batch(raw)
                    await self.bac.dsend(['OK'], z_route)
                    continue
                else:
                    await self.bac.dsend_err("data empty!", z_route)
                    continue

            elif sig == "chunk":
                try:
                    cmd = raw.pop(0)
                    db = raw.pop(0)
                    coll = raw.pop(0)
                    params = raw.pop(0)
                except Exception as e:
                    self.glog.error(f"Error while parsing Download: {e}")
                    await self.bac.dsend_err('Error parsing Download query.', z_route)
                    continue

                query = [cmd, db, coll, params]

                await resource_sock.psnd(query)
                records = await resource_sock.precv()

                await self.bac.dsend(records, z_route)
                continue

            elif sig == 'measurements':
                try:
                    cmd = raw.pop(0)
                    db = raw.pop(0)
                except Exception as e:
                    self.glog.error(f"Error parsing measurement query: {e}")
                    await self.bac.dsend_err('Error parsing measurement query', z_route)
                    continue
                await resource_sock.psnd(['measurements', db]) # here you can create loops by asking for it
                colls = await resource_sock.precv()
                await self.bac.dsend(colls, z_route)

            elif sig == 'count':
                try:
                    cmd = raw.pop(0)
                    db = raw.pop(0)
                    measurement = raw.pop(0)
                    query = [cmd, db, measurement]
                    if len(raw) == 1:
                        query.extend(raw)
                except Exception as e:
                    self.glog.error(f"Error parsing count query: {e}")
                    await self.bac.dsend_err('Error parsing count query', z_route)
                    continue
                await resource_sock.psnd(query)
                colls = await resource_sock.precv()
                await self.bac.dsend(colls, z_route)

            elif sig == "inventory":
                try:
                    cmd = raw.pop(0)
                    db = raw.pop(0)
                    tmp = raw.pop(0)
                    state = {}
                    state['inventory'] = {}
                    # TODO
                    state['store'] = {}
                except Exception as e:
                    self.glog.error(f"Error while parsing inventory: {e}")
                    await self.bac.dsend_err('Error parsing inventory query', z_route)
                    continue

                if tmp == 'all':
                    query = ['measurements', db]
                    await resource_sock.psnd(query)
                    colls = await resource_sock.precv()
                else:
                    colls = [tmp]

                for coll in colls:
                    state['inventory'][coll] = {}
                    try:
                        head_query = ['get', db, coll, 'head']
                        await resource_sock.psnd(head_query)
                        points = await resource_sock.precv()
                        state['inventory'][coll]['head'] = points.pop(0)

                        tail_query = ['get', db, coll, 'tail']
                        await resource_sock.psnd(tail_query)
                        points = await resource_sock.precv()
                        state['inventory'][coll]['tail'] = points.pop(0)
                    except Exception as e:
                        self.glog.error(f'Inventory update failed for {coll}: {e}')

                #print(state)
                await self.bac.dsend(state, z_route)
                continue

            else:
                await self.bac.dsend_err('Not so nice of you!', z_route)

def resource_worker_process(kcfg):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    kcfg['context'] = Kontext()
    kcfg['loop'] = loop

    data = KResourceWorker(**kcfg)
    data.start()



"""
    Copyright (c) 2019 Contributors as noted in the AUTHORS file
    This file is part of ggm, the GILGAMESH core engine in Python.
    ggm is free software; you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License (GPL) as published
    by the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
    As a special exception, the Contributors give you permission to link
    this library with independent modules to produce an executable,
    regardless of the license terms of these independent modules, and to
    copy and distribute the resulting executable under terms of your choice,
    provided that you also meet, for each linked independent module, the
    terms and conditions of the license of that module. An independent
    module is a module which is not derived from or based on this library.
    If you modify this library, you must extend this exception to your
    version of the library.
    ggm is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
    License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import zmq
import asyncio
from zmq.utils.strtypes import asbytes

class GHeartt(object):
    """
    heartbeater/handle_pong -> coroutines for hearbeater
    heart -> zmq thread device
    TODO write doc
    """
    def __init__(self, *args, **kwargs):

        self.HEART  = "ipc:///tmp/heart"
        self.BEAT   = "ipc:///tmp/beat"

        self.hearts = set()
        self.responses = set()
        self.period = 15 # *2 ??? # FIXME strange behaviour?
        self.lifetime = 0
        self.tic = self.loop.time()

    async def start_heart(self):
        """
        heart coroutine
        """
        ping = self.context.socket(zmq.SUB)
        ping.subscribe(b"")
        ping.connect(self.HEART)
        pong = self.context.socket(zmq.PUB)
        pong.connect(self.BEAT)
        while True:
            msg = await ping.recv_multipart()
            msg.insert(0, asbytes(self.name))
            pong.send_multipart(msg)

    async def handle_pong(self):
        """
        if heart is beating
        """
        pong = self.context.socket(zmq.SUB)
        pong.subscribe(b"")
        pong.bind(self.BEAT)
        while True:
            msg = await pong.recv_multipart()
            if float(msg[1]) == self.lifetime:
                self.responses.add(msg[0].decode())
            else:
                self.glog.error("{0} Bad Heartbeat from {1}: {2}".format(self.lifetime, msg[0], msg[1]))

    async def heartbeater(self):
        """
        beat the devil out of it
        """
        ping = self.context.socket(zmq.PUB)
        ping.bind(self.HEART)
        while True:
            await asyncio.sleep(self.period)

            toc = self.loop.time()
            self.lifetime += toc-self.tic
            self.tic = toc

            goodhearts = self.hearts.intersection(self.responses)
            heartfailures = self.hearts.difference(goodhearts)
            newhearts = self.responses.difference(goodhearts)
            #print(newhearts, goodhearts, heartfailures)
            #list(map(...)) needed, or map will not get *executed* as expected
            list(map(self.handle_new_heart, newhearts))
            list(map(self.handle_heart_failure, heartfailures))
            self.responses = set()

            #print("%i beating hearts: %s"%(len(self.hearts),self.hearts))
            ping.send_string(str(self.lifetime))

    def handle_new_heart(self, heart):
        self.hearts.add(heart)

    def handle_heart_failure(self, heart):
        """
        restarts stopped kernels
        """
        self.hearts.remove(heart)
        self.glog.warning(f'(Heart Failure) {self.period}s of inactivity from {heart}')

        for n, kernel in enumerate(self.kernel_map):
            if kernel['name'] == heart:
                self.kernel_map[n]['proc'].terminate()
                self.init_kernel(self.kernel_map[n], n)



import os
from setuptools import setup, find_packages

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = 'gilgamesh',
    version = '0.9.7',
    scripts=['gilgamesh'] ,
    author = 'Leonard Pollak',
    author_email = 'leonardp@tr-host.de',
    description = ('distributed data'),
    license = 'GPL-3',
    keywords = 'beta data',
    url = 'https://gitlab.com/gilgamesh-zmq/ggm',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Utilities',
        'Topic :: System :: Distributed Computing',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3.7',
    ],
    python_requires='>=3.7',
    install_requires=['pyzmq', 'aiohttp'],
    extras_require={
        'timescaledb': ['psycopg2'],
        'mqtt': ['paho-mqtt'],
    },
    packages=find_packages(exclude=['doc', 'config']),
    package_data = {
        'ggm': [
            'distfiles/*.json',
            'distfiles/*.service',
            'distfiles/*.dist',
            ],
    },
#    entry_points={
#        'console_scripts': [
#            'gilgamesh=gilgamesh:main',
#        ],
#    },
)



"""
    Copyright (c) 2019 Contributors as noted in the AUTHORS file
    This file is part of ggm, the GILGAMESH core engine in Python.
    ggm is free software; you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License (GPL) as published
    by the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
    As a special exception, the Contributors give you permission to link
    this library with independent modules to produce an executable,
    regardless of the license terms of these independent modules, and to
    copy and distribute the resulting executable under terms of your choice,
    provided that you also meet, for each linked independent module, the
    terms and conditions of the license of that module. An independent
    module is a module which is not derived from or based on this library.
    If you modify this library, you must extend this exception to your
    version of the library.
    ggm is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
    License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


def update_nested_dict(d, u):
    """doc
    """
    for k, v in u.items():
        dv = d.get(k, {})
        if not isinstance(dv, dict):
            d[k] = v
        elif isinstance(v, dict):
            d[k] = update_nested_dict(dv, v)
        else:
            d[k] = v
    return d

class Kmap(object):
    """FIXME
    there has to be a better way
    """
    def __init__(self, base_cfg, *args, **kwargs):
        self.base_cfg = base_cfg
        self.kmap = list()

    def append_spec(self, what, kname, config, xname=None, path=None):
        """doc
        """
        funcname = kname+'_process'
        modname = f'ggm.{what}.{kname}'

        if path:
            import importlib.util as iutil
            filename = kname+f'.py'
            f = path / filename
            spec = iutil.spec_from_file_location(modname, f)
            module = iutil.module_from_spec(spec)
            spec.loader.exec_module(module)
        else:
            from importlib import import_module
            module = import_module(modname)

        func = getattr(module, funcname)

        # use xname if available
        if 'config' not in config:
            config['config'] = dict()
        kname = xname or kname
        config['config']['name'] = kname
        config['config'].update(self.base_cfg)
        #update_nested_dict(config, self.base_cfg)
        kspec = {
            'name': kname,
            'function': func,
            'config': config['config']
        }
        self.kmap.append(kspec)
        return True

def generate_certificate(path, name):
    """FIXME
    there has to be a better way
    """
    import os, shutil, errno
    import zmq.auth
    from pathlib import Path

    if zmq.zmq_version_info() < (4,0):
        raise RuntimeError(f'Security is not supported in libzmq version < 4.0. libzmq version {zmq.zmq_version()}')

    # Generate client and server CURVE certificate files
    #base_dir = Path.cwd()
    base_dir = Path(path)
    tmp_dir = base_dir / 'certificates'

    if not base_dir.exists():
        os.mkdir(base_dir)
    if not tmp_dir.exists():
        os.mkdir(tmp_dir)

    # Check for existing certificates
    knames = [f'{name}.key', f'{name}.key_secret']
    for key in knames:
        f =  tmp_dir / key
        if f.exists():
            raise FileExistsError(f'{f} already exists.')

    # create new keys in certificate dir
    public_file, secret_file = zmq.auth.create_certificates(tmp_dir, name)

    pubk = f'{name}.key'
    pub_dir = path / Path(f'public_keys')
    seck = f'{name}.key_secret'
    sec_dir = path / Path(f'private_keys')
    if not pub_dir.exists():
        os.mkdir(pub_dir)
    if not sec_dir.exists():
        os.mkdir(sec_dir)

    dest = pub_dir / pubk
    if not dest.exists():
        shutil.move(public_file, dest)
    else:
        raise FileExistsError(f'{dest} already exists. trouble ahead...')

    dest = sec_dir / seck
    if not dest.exists():
        shutil.move(secret_file, dest)
    else:
        raise FileExistsError(f'{dest} already exists. trouble ahead...')

    try:
        os.rmdir(tmp_dir)
    except Exception as e:
        print(e)



"""
    Copyright (c) 2019 Contributors as noted in the AUTHORS file
    This file is part of ggm, the GILGAMESH core engine in Python.
    ggm is free software; you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License (GPL) as published
    by the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
    As a special exception, the Contributors give you permission to link
    this library with independent modules to produce an executable,
    regardless of the license terms of these independent modules, and to
    copy and distribute the resulting executable under terms of your choice,
    provided that you also meet, for each linked independent module, the
    terms and conditions of the license of that module. An independent
    module is a module which is not derived from or based on this library.
    If you modify this library, you must extend this exception to your
    version of the library.
    ggm is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
    License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import zlib
import json
import pickle
import asyncio
import zmq.asyncio

class GSock(zmq.asyncio.Socket):
    """A class with some extra serialization methods
        this is the place for the nasty protocol code
        roll your is advised
    """
    async def psnd(self, msg, compression=False):
        """
        send msg with internal routing as list
        and pickled payload
        """
        # if msg is dict -- return immediately
        # if msg len is 1 -- return immediately
        if isinstance(msg, dict):
            ret = await self.send_pyobj(msg)
            return ret
        elif not isinstance(msg, list):
            return True # !!! FIXME
        elif len(msg) <= 1:
            ret = await self.send_pyobj(msg)
            return ret

        p = []
        p.extend(msg[:-1])
        if compression:
            comp = zlib.compress(pickle.dumps(msg[-1]))
            p.append(comp)
        else:
            p.append(pickle.dumps(msg[-1]))
        ret = await self.send_pyobj(p)
        return ret

    async def precv(self, compression=False):
        """
        recv msg with internal routing as list
        and unpickle payload
        """
        msg = await self.recv_pyobj()
        if isinstance(msg, dict):
            return msg
        elif not isinstance(msg, list):
            return True # !!! FIXME
        elif len(msg) <= 1:
            return msg

        route = msg[:-1]
        if compression:
            decomp = pickle.loads(zlib.decompress(msg[-1]))
            msg[-1] = decomp
        else:
            msg[-1] = pickle.loads(msg[-1])

        return msg

    async def dsend(self, msg, z_route=None, compression=False):
        """
        sending as dealer to dealer (resource worker/client/...)

        emtpy frame for compatibility with req/rep OR zmq internal route
        """
        p = z_route or [b'']

        # if msg is dict return immediately
        if isinstance(msg, dict):
            p.append(pickle.dumps(msg))
            ret = await self.send_multipart(p)
            return ret
        elif not isinstance(msg, list):
            return True # !!! FIXME
        elif len(msg) <= 1:
            p.append(pickle.dumps(msg))
            ret = await self.send_multipart(p)
            return ret

        # internal app route
        payload = msg[:-1]

        if compression:
            comp = zlib.compress(pickle.dumps(msg[-1]))
            payload.append(comp)
        else:
            payload.append(pickle.dumps(msg[-1]))

        p.append(pickle.dumps(payload))
        ret = await self.send_multipart(p)
        return ret

    async def drecv(self, compression=False):
        """receiving as dealer from dealer(resource worker)
        compatible with req/rep
        """
        rawmsg = await self.recv_multipart()
        route = rawmsg[:-1]
        msg = pickle.loads(rawmsg[-1])
        if isinstance(msg, dict):
            return route, msg
        elif not isinstance(msg, list):
            return True # !!! FIXME
        elif len(msg) <= 1:
            return route, msg

        if compression:
            decomp = pickle.loads(zlib.decompress(msg[-1]))
            msg[-1] = decomp
        else:
             msg[-1] = pickle.loads(msg[-1])
        #print(route, msg)
        return route, msg

    async def upstream(self, cmd, segment, source, compression=False):

        CHUNK_SIZE = 1000
        PIPELINE = 1

        credit = PIPELINE   # Up to PIPELINE chunks in transit
        total = 0           # Total records received
        chunks = 0          # Total chunks received
        offset = 0          # Offset of next chunk request

        up_cmd = ['series', 'upload']
        up_cmd.extend(cmd[2:])
        up_cmd.append(b'') #FIXME WTF

        cmd.append(segment)

        while True:
            while credit:
                segment['offset'] = offset
                segment['limit']= CHUNK_SIZE
                cmd[-1] = segment
                await source.dsend(cmd)
                offset += CHUNK_SIZE
                credit -= 1
            route, msg = await source.drecv(compression)
            up_cmd[-1] = msg
            if msg:
                await self.dsend(up_cmd)
                route, reply = await self.drecv()
            chunks += 1
            credit += 1
            size = len(msg)
            if size < CHUNK_SIZE:
                break
            # FIXME hack for upload 'flow control'
            await asyncio.sleep(3)

        # 'empty' pipeline
        while credit < PIPELINE:
            route, msg = await source.drecv(compression)
            up_cmd[-1] = msg
            if msg:
                await self.dsend(up_cmd)
                route, reply = await self.drecv()
            credit += 1
        #!!
        return await asyncio.sleep(0.1)

    async def psend_noerr(self, msg):
        await self.psnd({"Error": False, 'Reason': msg})
    async def dsend_noerr(self, msg, z_route=None):
        await self.dsend({"Error": False, 'Reason': msg}, z_route)
    async def psend_err(self, errmsg):
        await self.psnd({"Error": True, 'Reason': errmsg})
    async def dsend_err(self, errmsg, z_route=None):
        await self.dsend({"Error": True, 'Reason': errmsg}, z_route)

class Kontext(zmq.asyncio.Context):
    _socket_class = GSock



import sys
import zlib
import json
import pickle
import asyncio
import zmq.auth
import time
import socket
from .sync_context import Kontext

from pathlib import Path
from datetime import datetime as dt

from pprint import pprint

import pandas as pd
from pandas.io.json import json_normalize

class GClient(object):
    def __init__(self, ggm_host='local',*args, **kwargs):
        """GILGAMESH™
        
        class GGM(ggm_host='local')
        
        ggm_host: sting found as key in config.json
        """
        cfg_file = Path.home()/'.gilgamesh'/'config'/'client_config.json'
        if not cfg_file.exists():
            raise FileNotFoundError(f'{cfg_file} does not exist.')
        
        with open(cfg_file, 'r') as raw:
            cfg = json.load(raw)
        
        if not ggm_host in cfg:
            raise ValueError(f'{ggm_host} not found in config.')

        cfg = cfg[ggm_host]
        
        self.ctx = Kontext()
        self.client = self.ctx.socket(zmq.DEALER)
        
        keys_path = Path.home()/'.gilgamesh'/'keys'
        client_secret_file = keys_path/'private_keys'/f'{cfg["client_key"]}.key_secret'
        server_public_file = keys_path/'public_keys'/f'{cfg["server_key"]}.key'
        
        client_public, client_secret = zmq.auth.load_certificate(client_secret_file)
        server_public, _ = zmq.auth.load_certificate(server_public_file)
        
        self.client.curve_secretkey = client_secret
        self.client.curve_publickey = client_public
        self.client.curve_serverkey = server_public
        
        if not 'ip' in cfg:
            try:
                ip = socket.gethostbyname(cfg['hostname'])
                cfg['ip'] = ip
            except:
                raise ConnectionError(f'Could not resolve {hostname}')
        self.client.connect(f'tcp://{cfg["ip"]}:{cfg["port"]}')
        
        self.poller = zmq.Poller()
        self.poller.register(self.client, zmq.POLLIN)

        assert self.check_conn()
        
        self.client.psend(['info'])
        reply = self.client.precv()
        self.remote_dev_id = reply['dev_id']
        
    def show_devices(self):
        self.client.psend(['json', 'get', 'device_state_db', self.remote_dev_id, 'head'])
        reply = self.client.precv()
        devs = [d for d in reply['device_state_db'][self.remote_dev_id]['store'].keys()]
        devs.append(self.remote_dev_id)
        return devs

    def show_measurements(self, device):
        self.client.psend(['json', 'get', 'device_state_db', self.remote_dev_id, 'head'])
        reply = self.client.precv()
        if device == self.remote_dev_id:
            return [d for d in reply['device_state_db'][self.remote_dev_id]['inventory'].keys()]
        else:
            return [d for d in reply['device_state_db'][self.remote_dev_id]['store'][device].keys()]
    
    def show_store(self):
        self.client.psend(['json', 'get', 'device_state_db', self.remote_dev_id, 'head'])
        reply = self.client.precv()
        store = reply['device_state_db'][self.remote_dev_id]['store']
        store[self.remote_dev_id] = reply['device_state_db'][self.remote_dev_id]['inventory']
        return store

    def get_measurement(self, dev_id, measurement, start=None, stop=None, chunk_size=5000, cols=None, pipeline=3, compression=False, progress=True):
        segment = dict()
        if not start:
            return 'error: start must be given'
        else:
            segment['start'] = start

        segment['stop'] = stop or self.get_tmst_now()
        if cols:
            segment['cols'] = cols

        if progress:
            count_cmd = ['series', 'count', dev_id, measurement, segment]
            self.client.psend(count_cmd)
            raw = self.client.precv()
            count_all = raw.pop()
            total_chunks = int(count_all/chunk_size)

        cmd = ['series', 'chunk', dev_id, measurement]

        try:
            start = time.time()
            if progress:
                data = self.download_df(cmd, segment, chunk_size=chunk_size, pipeline=pipeline, compression=False, total_chunks=total_chunks)
            else:
                data = self.download_df(cmd, segment, chunk_size=chunk_size, pipeline=pipeline, compression=False)
            stop = time.time()
            print('\n')
            print(f'-- {dev_id} {measurement} --')
            print(f'{len(data)} records received.')
            print(f'Time elapsed: {round(stop-start, 2)} s')
            print('\n')
        except Exception as e:
            print(f'Error while downloading {dev_id} {measurement} ({e})')
            data = self.clean_pipeline()
        except KeyboardInterrupt:
            data = self.clean_pipeline()

        return data

    def get_tmst_now(self):
        return dt.utcnow().isoformat(sep=' ')

    def clean_pipeline(self):
        msg = []
        while len(dict(self.poller.poll(1000))) > 0:
            _, msg = self.client.drecv()
        if isinstance(msg, dict):
            if 'Reason' in msg.keys():
                pprint(msg['Reason'])
        return True

    def update_progress(self, chunk, total):
        sys.stdout.write(f'\rDownloading chunk {chunk}/{total}')
        sys.stdout.flush()

    def download_df(self, cmd, segment, chunk_size=5000, pipeline=3, compression=False, total_chunks=None):

        CHUNK_SIZE = chunk_size
        PIPELINE = pipeline

        credit = PIPELINE    # Up to PIPELINE chunks in transit
        total = 0            # Total records received
        chunks = 0           # Total chunks received
        offset = 0           # Offset of next chunk request
        thed = pd.DataFrame()# zis is ze data
        cmd.append(segment)

        while True:
            while credit:
                segment['offset'] = offset
                segment['limit']= CHUNK_SIZE
                cmd[-1] = segment
                self.client.dsend(cmd)
                offset += CHUNK_SIZE
                credit -= 1
            route, msg = self.client.drecv(compression)
            if len(msg) > 0:
                tmp_df = json_normalize(msg)
                tmp_df = tmp_df.set_index('time')
                tmp_df.index = pd.to_datetime(tmp_df.index)
                thed = thed.append(tmp_df)
            if total_chunks:
                self.update_progress(chunks, total_chunks)
            chunks += 1
            credit += 1
            size = len(msg)
            if size < CHUNK_SIZE:
                break

        # 'empty' pipeline
        while credit < PIPELINE:
            route, msg = self.client.drecv(compression)
            credit += 1
            if len(msg) > 0:
                tmp_df = json_normalize(msg)
                tmp_df = tmp_df.set_index('time')
                tmp_df.index = pd.to_datetime(tmp_df.index)
                thed = thed.append(tmp_df)
        #remove the prefix for the column names
        thed.rename(columns={k:k.replace('fields.','').replace('tags.','') for k in list(thed.keys())}, inplace=True)
        
        return thed

    def down_raw(self, dev_id, measurement, start=None, stop=None, chunk_size=5000, cols=None, pipeline=3, compression=False):
        segment = dict()
        if not start:
            return 'error: start must be given'
        else:
            segment['start'] = start

        segment['stop'] = stop or self.get_tmst_now()
        if cols:
            segment['cols'] = cols

        cmd = ['series', 'chunk', dev_id, measurement]
        CHUNK_SIZE = chunk_size
        PIPELINE = pipeline

        credit = PIPELINE    # Up to PIPELINE chunks in transit
        total = 0            # Total records received
        chunks = 0           # Total chunks received
        offset = 0           # Offset of next chunk request
        thed = []# zis is ze data
        cmd.append(segment)

        while True:
            while credit:
                segment['offset'] = offset
                segment['limit']= CHUNK_SIZE
                cmd[-1] = segment
                self.client.dsend(cmd)
                offset += CHUNK_SIZE
                credit -= 1
            route, msg = self.client.drecv(compression)
            if len(msg) > 0:
                thed.append(msg)
            chunks += 1
            credit += 1
            size = len(msg)
            if size < CHUNK_SIZE:
                break

        # 'empty' pipeline
        while credit < PIPELINE:
            route, msg = self.client.drecv(compression)
            credit += 1
            if len(msg) > 0:
                thed.append(msg)

        return thed.pop()

    def check_conn(self):        
        self.client.psend(['greetings'])
        while True:
            socks = dict(self.poller.poll(1000))
            if socks.get(self.client) == zmq.POLLIN:
                ret = self.client.precv()
                if ret == ['earthlings']:
                    print('Connecting to gilgamesh successful!')
                    return True
                elif not ret == ['earthlings']:
                    print(f'failed greeting(!) got: {ret}\nretrying...')
                    self.client.psend(['greetings'])
                    continue
            elif not socks.get(self.client):
                self.terminate()
                print(f'Failed connecting to server please check settings!\n')
                return False

    def terminate(self):
        self.client.set(zmq.LINGER, 0)
        self.client.close()
        self.ctx.term()
        print(f'Terminated gilgamesh Client')
